__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
import unittest

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
from lenstronomy.LightModel.Profiles.starlets import SLIT_Starlets


_force_no_pysap = (
    True  # if issues on Travis-CI to install pysap, force use python-only functions
)


class TestImageModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)
        kernel = psf_class.kernel_point_source
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel,
            "psf_error_map": np.ones_like(kernel) * 0.001,
        }
        psf_class = PSF(**kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        kwargs_lens_light_base = [kwargs_sersic]
        lens_light_model_class_base = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source_base = [kwargs_sersic_ellipse]
        source_model_class_base = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.01, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class_base = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_numerics_base = {
            "supersampling_factor": 2,
            "supersampling_convolution": False,
        }
        imageModel_base = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class_base,
            lens_light_model_class_base,
            point_source_class_base,
            kwargs_numerics=kwargs_numerics_base,
        )
        image_sim = sim_util.simulate_simple(
            imageModel_base,
            self.kwargs_lens,
            kwargs_source_base,
            kwargs_lens_light_base,
            self.kwargs_ps,
        )
        data_class.update_data(image_sim)

        # create a starlet light distributions
        n_scales = 6
        source_map = imageModel_base.source_surface_brightness(
            kwargs_source_base, de_lensed=True, unconvolved=True
        )
        starlets_class = SLIT_Starlets(force_no_pysap=_force_no_pysap)
        source_map_starlets = starlets_class.decomposition_2d(source_map, n_scales)
        self.kwargs_source = [
            {
                "amp": source_map_starlets,
                "n_scales": n_scales,
                "n_pixels": numPix,
                "scale": deltaPix,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        source_model_class = LightModel(light_model_list=["SLIT_STARLETS"])
        lens_light_map = imageModel_base.lens_surface_brightness(
            kwargs_lens_light_base, unconvolved=True
        )
        starlets_class = SLIT_Starlets(force_no_pysap=_force_no_pysap, second_gen=True)
        lens_light_starlets = starlets_class.decomposition_2d(lens_light_map, n_scales)
        self.kwargs_lens_light = [
            {
                "amp": lens_light_starlets,
                "n_scales": n_scales,
                "n_pixels": numPix,
                "scale": deltaPix,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        lens_light_model_class = LightModel(light_model_list=["SLIT_STARLETS_GEN2"])

        kwargs_numerics = {"supersampling_factor": 1}
        kwargs_pixelbased = {
            "supersampling_factor_source": 2,  # supersampling of pixelated source grid
            # following choices are to minimize pixel solver runtime (not to get accurate reconstruction!)
            "threshold_decrease_type": "none",
            "num_iter_source": 2,
            "num_iter_lens": 2,
            "num_iter_global": 2,
            "num_iter_weights": 2,
        }
        self.imageModel = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            point_source_class=None,
            kwargs_numerics=kwargs_numerics,
            kwargs_pixelbased=kwargs_pixelbased,
        )
        self.imageModel_source = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=None,
            point_source_class=None,
            kwargs_numerics=kwargs_numerics,
            kwargs_pixelbased=kwargs_pixelbased,
        )

        self.solver = LensEquationSolver(lensModel=self.imageModel.LensModel)

    def test_source_surface_brightness(self):
        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=False, de_lensed=True
        )
        assert len(source_model) == 100

        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=False, de_lensed=False
        )
        assert len(source_model) == 100
        npt.assert_almost_equal(
            source_model[10, 10], 0.13939841209844345 * 0.05**2, decimal=4
        )

        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=True, de_lensed=False
        )
        assert len(source_model) == 100
        npt.assert_almost_equal(
            source_model[10, 10], 0.13536114618182182 * 0.05**2, decimal=4
        )

    def test_lens_surface_brightness(self):
        lens_flux = self.imageModel.lens_surface_brightness(
            self.kwargs_lens_light, unconvolved=False
        )
        npt.assert_almost_equal(lens_flux[50, 50], 0.011827638016863616, decimal=4)

        # the following should raise an error: no deconvolution is performed when pixel-based modelling lens light
        # lens_flux = self.imageModel.lens_surface_brightness(self.kwargs_lens_light, unconvolved=True)

    def test_image_linear_solve(self):
        model, error_map, cov_param, param = self.imageModel.image_linear_solve(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        chi2_reduced = self.imageModel.reduced_chi2(model, error_map)
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)

        model, error_map, cov_param, param = self.imageModel_source.image_linear_solve(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        chi2_reduced = self.imageModel.reduced_chi2(model, error_map)
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)

    def test_image_with_params(self):
        model = self.imageModel.image(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            unconvolved=False,
            source_add=True,
            lens_light_add=True,
            point_source_add=True,
        )
        error_map = self.imageModel._error_map_psf(self.kwargs_lens, self.kwargs_ps)
        chi2_reduced = self.imageModel.reduced_chi2(model, error_map)
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)

    def test_likelihood_data_given_model(self):
        logL = self.imageModel.likelihood_data_given_model(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            source_marg=False,
        )
        npt.assert_almost_equal(logL, -5000, decimal=-3)

    def test_reduced_residuals(self):
        model = sim_util.simulate_simple(
            self.imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            no_noise=True,
        )
        residuals = self.imageModel.reduced_residuals(model, error_map=0)
        npt.assert_almost_equal(np.std(residuals), 1.01, decimal=1)

        chi2 = self.imageModel.reduced_chi2(model, error_map=0)
        npt.assert_almost_equal(chi2, 1, decimal=1)

    def test_numData_evaluate(self):
        numData = self.imageModel.num_data_evaluate
        assert numData == 10000

    def test_num_param_linear(self):
        num_param_linear = self.imageModel.num_param_linear(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        assert (
            num_param_linear == 0
        )  # pixels of pixel-based profiles not counted as linear param

        num_param_linear = self.imageModel_source.num_param_linear(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        assert (
            num_param_linear == 0
        )  # pixels of pixel-based profiles not counted as linear param

    def test_update_data(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        self.imageModel.update_data(data_class)
        assert self.imageModel.Data.num_pixel == 100

    def test_create_empty(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)
        imageModel_empty = ImageModel(data_class, PSF())
        assert imageModel_empty._psf_error_map == False

        flux = imageModel_empty.lens_surface_brightness(kwargs_lens_light=None)
        assert flux.all() == 0

    def test_extinction_map(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)
        extinction_class = DifferentialExtinction(
            optical_depth_model=["UNIFORM"], tau0_index=0
        )
        imageModel = ImageModel(data_class, PSF(), extinction_class=extinction_class)
        extinction = imageModel.extinction_map(
            kwargs_extinction=[{"amp": 1}], kwargs_special={"tau0_list": [1, 0, 0]}
        )
        npt.assert_almost_equal(extinction, np.exp(-1))

    def test_error_response(self):
        C_D_response, psf_model_error = self.imageModel._error_response(
            self.kwargs_lens, self.kwargs_ps, kwargs_special=None
        )
        assert len(psf_model_error) == 100
        print(np.sum(psf_model_error))
        npt.assert_almost_equal(np.sum(psf_model_error), 0, decimal=3)

        C_D_response, psf_model_error = self.imageModel_source._error_response(
            self.kwargs_lens, self.kwargs_ps, kwargs_special=None
        )
        assert len(psf_model_error) == 100
        print(np.sum(psf_model_error))
        npt.assert_almost_equal(np.sum(psf_model_error), 0, decimal=3)


class TestRaise(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestRaise, self).__init__(*args, **kwargs)
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg, inverse=True
        )
        self.data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)
        kernel = psf_class.kernel_point_source
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel,
            "psf_error_map": np.ones_like(kernel) * 0.001,
        }
        self.psf_class = PSF(**kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        self.lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        kwargs_lens_light_base = [kwargs_sersic]
        lens_light_model_class_base = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        kwargs_source_base = [kwargs_sersic_ellipse]
        source_model_class_base = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.01, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class_base = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_numerics_base = {
            "supersampling_factor": 2,
            "supersampling_convolution": False,
        }
        imageModel_base = ImageModel(
            self.data_class,
            self.psf_class,
            self.lens_model_class,
            source_model_class_base,
            lens_light_model_class_base,
            point_source_class_base,
            kwargs_numerics=kwargs_numerics_base,
        )
        image_sim = sim_util.simulate_simple(
            imageModel_base,
            self.kwargs_lens,
            kwargs_source_base,
            kwargs_lens_light_base,
            self.kwargs_ps,
        )
        self.data_class.update_data(image_sim)

        # create a starlet light distributions
        n_scales = 6
        source_map = imageModel_base.source_surface_brightness(
            kwargs_source_base, de_lensed=True, unconvolved=True
        )
        starlets_class = SLIT_Starlets(force_no_pysap=_force_no_pysap)
        source_map_starlets = starlets_class.decomposition_2d(source_map, n_scales)
        self.kwargs_source = [
            {
                "amp": source_map_starlets,
                "n_scales": n_scales,
                "n_pixels": numPix,
                "scale": deltaPix,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        self.source_model_class = LightModel(light_model_list=["SLIT_STARLETS"])
        lens_light_map = imageModel_base.lens_surface_brightness(
            kwargs_lens_light_base, unconvolved=True
        )
        starlets_class = SLIT_Starlets(force_no_pysap=_force_no_pysap, second_gen=True)
        lens_light_starlets = starlets_class.decomposition_2d(lens_light_map, n_scales)
        self.kwargs_lens_light = [
            {
                "amp": lens_light_starlets,
                "n_scales": n_scales,
                "n_pixels": numPix,
                "scale": deltaPix,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        self.lens_light_model_class = LightModel(
            light_model_list=["SLIT_STARLETS_GEN2"]
        )

        self.kwargs_numerics = {"supersampling_factor": 1}
        self.kwargs_pixelbased = {
            "supersampling_factor_source": 2,  # supersampling of pixelated source grid
            # following choices are to minimize pixel solver runtime (not to get accurate reconstruction!)
            "threshold_decrease_type": "none",
            "num_iter_source": 2,
            "num_iter_lens": 2,
            "num_iter_global": 2,
            "num_iter_weights": 2,
        }

    def test_raise(self):
        with self.assertRaises(ValueError):
            # test various numerics that are not supported by the pixelbased solver
            kwargs_numerics = {
                "supersampling_factor": 2,
                "supersampling_convolution": True,
            }
            imageModel = ImageLinearFit(
                self.data_class,
                self.psf_class,
                self.lens_model_class,
                source_model_class=self.source_model_class,
                lens_light_model_class=self.lens_light_model_class,
                kwargs_numerics=kwargs_numerics,
                kwargs_pixelbased=self.kwargs_pixelbased,
            )
        with self.assertRaises(ValueError):
            # test various numerics that are not supported by the pixelbased solver
            kwargs_numerics = {"compute_mode": "adaptive"}
            imageModel = ImageLinearFit(
                self.data_class,
                self.psf_class,
                self.lens_model_class,
                source_model_class=self.source_model_class,
                lens_light_model_class=self.lens_light_model_class,
                kwargs_numerics=kwargs_numerics,
                kwargs_pixelbased=self.kwargs_pixelbased,
            )
        with self.assertRaises(ValueError):
            # test unsupported gaussian PSF type
            kwargs_psf = {
                "psf_type": "GAUSSIAN",
                "fwhm": 0.5,
                "truncation": 5,
                "pixel_size": 0.05,
            }
            psf_class = PSF(**kwargs_psf)
            imageModel = ImageLinearFit(
                self.data_class,
                psf_class,
                self.lens_model_class,
                source_model_class=self.source_model_class,
                lens_light_model_class=self.lens_light_model_class,
                kwargs_numerics=self.kwargs_numerics,
                kwargs_pixelbased=self.kwargs_pixelbased,
            )
        with self.assertRaises(ValueError):
            kwargs_numerics = {"supersampling_factor": 1}
            # test more than a single pixel-based light profile
            source_model_class = LightModel(["SLIT_STARLETS", "SLIT_STARLETS"])
            imageModel = ImageLinearFit(
                self.data_class,
                self.psf_class,
                self.lens_model_class,
                source_model_class=source_model_class,
                lens_light_model_class=self.lens_light_model_class,
                kwargs_numerics=self.kwargs_numerics,
                kwargs_pixelbased=self.kwargs_pixelbased,
            )
        with self.assertRaises(ValueError):
            # test access to unconvolved lens light surface brightness
            imageModel = ImageLinearFit(
                self.data_class,
                self.psf_class,
                self.lens_model_class,
                source_model_class=self.source_model_class,
                lens_light_model_class=self.lens_light_model_class,
                kwargs_numerics=self.kwargs_numerics,
                kwargs_pixelbased=self.kwargs_pixelbased,
            )
            imageModel.lens_surface_brightness(self.kwargs_lens_light, unconvolved=True)


if __name__ == "__main__":
    pytest.main()



__author__ = "Simon Birrer"
"""This file contains standard routines."""

import numpy as np
import itertools
from lenstronomy.Util.numba_util import jit
from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def merge_dicts(*dict_args):
    """Given any number of dicts, shallow copy and merge into a new dict, precedence
    goes to key value pairs in latter dicts."""
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


@export
def approx_theta_E(ximg, yimg):
    dis = []
    xinds, yinds = [0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3]

    for i, j in zip(xinds, yinds):
        dx, dy = ximg[i] - ximg[j], yimg[i] - yimg[j]
        dr = (dx**2 + dy**2) ** 0.5
        dis.append(dr)
    dis = np.array(dis)

    greatest = np.argmax(dis)
    dr_greatest = dis[greatest]
    dis[greatest] = 0

    second_greatest = np.argmax(dis)
    dr_second = dis[second_greatest]

    return 0.5 * (dr_greatest * dr_second) ** 0.5


@export
def sort_image_index(ximg, yimg, xref, yref):
    """

    :param ximg: x coordinates to sort
    :param yimg: y coordinates to sort
    :param xref: reference x coordinate
    :param yref: reference y coordinate
    :return: indexes such that ximg[indexes],yimg[indexes] matches xref,yref
    """

    assert len(xref) == len(ximg)
    ximg, yimg = np.array(ximg), np.array(yimg)
    x_self = np.array(list(itertools.permutations(ximg)))
    y_self = np.array(list(itertools.permutations(yimg)))

    indexes = [0, 1, 2, 3]
    index_iterations = list(itertools.permutations(indexes))
    delta_r = []

    for i in range(0, int(len(x_self))):
        dr = 0
        for j in range(0, int(len(x_self[0]))):
            dr += (x_self[i][j] - xref[j]) ** 2 + (y_self[i][j] - yref[j]) ** 2

        delta_r.append(dr**0.5)

    min_indexes = np.array(index_iterations[np.argmin(delta_r)])

    return min_indexes


@export
@jit()
def rotate(xcoords, ycoords, angle):
    """

    :param xcoords: x points
    :param ycoords: y points
    :param angle: angle in radians
    :return: x points and y points rotated ccw by angle theta
    """
    return xcoords * np.cos(angle) + ycoords * np.sin(angle), -xcoords * np.sin(
        angle
    ) + ycoords * np.cos(angle)


@export
def map_coord2pix(ra, dec, x_0, y_0, M):
    """This routines performs a linear transformation between two coordinate systems.
    Mainly used to transform angular into pixel coordinates in an image.

    :param ra: ra coordinates
    :param dec: dec coordinates
    :param x_0: pixel value in x-axis of ra,dec = 0,0
    :param y_0: pixel value in y-axis of ra,dec = 0,0
    :param M: 2x2 matrix to transform angular to pixel coordinates
    :return: transformed coordinate systems of input ra and dec
    """
    x, y = M.dot(np.array([ra, dec]))
    return x + x_0, y + y_0


@export
def array2image(array, nx=0, ny=0):
    """Returns the information contained in a 1d array into an n*n 2d array (only works
    when length of array is n**2, or nx and ny are provided)

    :param array: image values
    :type array: array of size n**2
    :returns: 2d array
    :raises: AttributeError, KeyError
    """
    if nx == 0 or ny == 0:
        n = int(np.sqrt(len(array)))
        if n**2 != len(array):
            raise ValueError(
                "lenght of input array given as %s is not square of integer number!"
                % (len(array))
            )
        nx, ny = n, n
    image = array.reshape(int(nx), int(ny))
    return image


@export
def image2array(image):
    """Returns the information contained in a 2d array into an n*n 1d array.

    :param image: image values
    :type image: array of size (n,n)
    :returns: 1d array
    :raises: AttributeError, KeyError
    """
    nx, ny = image.shape  # find the size of the array
    imgh = np.reshape(image, nx * ny)  # change the shape to be 1d
    return imgh


@export
def array2cube(array, n_1, n_23):
    """Returns the information contained in a 1d array of shape (n_1*n_23*n_23) into 3d
    array with shape (n_1, sqrt(n_23), sqrt(n_23))

    :param array: image values
    :type array: 1d array
    :param n_1: first dimension of returned array
    :type n_1: int
    :param n_23: square of second and third dimensions of returned array
    :type n_23: int
    :returns: 3d array
    :raises ValueError: when n_23 is not a perfect square
    """
    n = int(np.sqrt(n_23))
    if n**2 != n_23:
        raise ValueError(
            "2nd and 3rd dims (%s) are not square of integer number!" % n_23
        )
    n_2, n_3 = n, n
    cube = array.reshape(n_1, n_2, n_3)
    return cube


@export
def cube2array(cube):
    """Returns the information contained in a 3d array of shape (n_1, n_2, n_3) into 1d
    array with shape (n_1*n_2*n_3)

    :param cube: image values
    :type cube: 3d array
    :returns: 1d array
    """
    n_1, n_2, n_3 = cube.shape
    array = cube.reshape(n_1 * n_2 * n_3)
    return array


@export
def make_grid(numPix, deltapix, subgrid_res=1, left_lower=False):
    """Creates pixel grid (in 1d arrays of x- and y- positions) default coordinate frame
    is such that (0,0) is in the center of the coordinate grid.

    :param numPix: number of pixels per axis Give an integers for a square grid, or a
        2-length sequence (first, second axis length) for a non-square grid.
    :param deltapix: pixel size
    :param subgrid_res: sub-pixel resolution (default=1)
    :return: x, y position information in two 1d arrays
    """

    # Check numPix is an integer, or 2-sequence of integers
    if isinstance(numPix, (tuple, list, np.ndarray)):
        assert len(numPix) == 2
        if any(x != round(x) for x in numPix):
            raise ValueError("numPix contains non-integers: %s" % numPix)
        numPix = np.asarray(numPix, dtype=int)
    else:
        if numPix != round(numPix):
            raise ValueError("Attempt to specify non-int numPix: %s" % numPix)
        numPix = np.array([numPix, numPix], dtype=int)

    # Super-resolution sampling
    numPix_eff = (numPix * subgrid_res).astype(int)
    deltapix_eff = deltapix / float(subgrid_res)

    # Compute unshifted grids.
    # X values change quickly, Y values are repeated many times
    x_grid = np.tile(np.arange(numPix_eff[0]), numPix_eff[1]) * deltapix_eff
    y_grid = np.repeat(np.arange(numPix_eff[1]), numPix_eff[0]) * deltapix_eff

    if left_lower is True:
        # Shift so (0, 0) is in the "lower left"
        # Note this does not shift when subgrid_res = 1
        shift = -1.0 / 2 + 1.0 / (2 * subgrid_res) * np.array([1, 1])
    else:
        # Shift so (0, 0) is centered
        shift = deltapix_eff * (numPix_eff - 1) / 2

    return x_grid - shift[0], y_grid - shift[1]


@export
def make_grid_transformed(numPix, Mpix2Angle):
    """Returns grid with linear transformation (deltaPix and rotation)

    :param numPix: number of Pixels
    :param Mpix2Angle: 2-by-2 matrix to mat a pixel to a coordinate
    :return: coordinate grid
    """
    x_grid, y_grid = make_grid(numPix, deltapix=1)
    ra_grid, dec_grid = map_coord2pix(x_grid, y_grid, 0, 0, Mpix2Angle)
    return ra_grid, dec_grid


@export
def make_grid_with_coordtransform(
    numPix,
    deltapix,
    subgrid_res=1,
    center_ra=0,
    center_dec=0,
    left_lower=False,
    inverse=True,
):
    """Same as make_grid routine, but returns the transformation matrix and shift
    between coordinates and pixel.

    :param numPix: number of pixels per axis
    :param deltapix: pixel scale per axis
    :param subgrid_res: super-sampling resolution relative to the stated pixel size
    :param center_ra: center of the grid
    :param center_dec: center of the grid
    :param left_lower: sets the zero point at the lower left corner of the pixels
    :param inverse: bool, if true sets East as left, otherwise East is righrt
    :return: ra_grid, dec_grid, ra_at_xy_0, dec_at_xy_0, x_at_radec_0, y_at_radec_0,
        Mpix2coord, Mcoord2pix
    """
    numPix_eff = numPix * subgrid_res
    deltapix_eff = deltapix / float(subgrid_res)
    a = np.arange(numPix_eff)
    matrix = np.dstack(np.meshgrid(a, a)).reshape(-1, 2)
    if inverse is True:
        delta_x = -deltapix_eff
    else:
        delta_x = deltapix_eff
    if left_lower is True:
        ra_grid = matrix[:, 0] * delta_x
        dec_grid = matrix[:, 1] * deltapix_eff
    else:
        ra_grid = (matrix[:, 0] - (numPix_eff - 1) / 2.0) * delta_x
        dec_grid = (matrix[:, 1] - (numPix_eff - 1) / 2.0) * deltapix_eff
    shift = (subgrid_res - 1) / (2.0 * subgrid_res) * deltapix
    ra_grid += -shift + center_ra
    dec_grid += -shift + center_dec
    ra_at_xy_0 = ra_grid[0]
    dec_at_xy_0 = dec_grid[0]

    Mpix2coord = np.array([[delta_x, 0], [0, deltapix_eff]])
    Mcoord2pix = np.linalg.inv(Mpix2coord)
    x_at_radec_0, y_at_radec_0 = map_coord2pix(
        -ra_at_xy_0, -dec_at_xy_0, x_0=0, y_0=0, M=Mcoord2pix
    )
    return (
        ra_grid,
        dec_grid,
        ra_at_xy_0,
        dec_at_xy_0,
        x_at_radec_0,
        y_at_radec_0,
        Mpix2coord,
        Mcoord2pix,
    )


@export
def grid_from_coordinate_transform(nx, ny, Mpix2coord, ra_at_xy_0, dec_at_xy_0):
    """Return a grid in x and y coordinates that satisfy the coordinate system.

    :param nx: number of pixels in x-axis
    :param ny: number of pixels in y-axis
    :param Mpix2coord: transformation matrix (2x2) of pixels into coordinate
        displacements
    :param ra_at_xy_0: RA coordinate at (x,y) = (0,0)
    :param dec_at_xy_0: DEC coordinate at (x,y) = (0,0)
    :return: RA coordinate grid, DEC coordinate grid
    """
    a = np.arange(nx)
    b = np.arange(ny)
    matrix = np.dstack(np.meshgrid(a, b)).reshape(-1, 2)
    x_grid = matrix[:, 0]
    y_grid = matrix[:, 1]
    ra_grid = x_grid * Mpix2coord[0, 0] + y_grid * Mpix2coord[0, 1] + ra_at_xy_0
    dec_grid = x_grid * Mpix2coord[1, 0] + y_grid * Mpix2coord[1, 1] + dec_at_xy_0
    return ra_grid, dec_grid


@export
def get_axes(x, y):
    """Computes the axis x and y of a given 2d grid.

    :param x:
    :param y:
    :return:
    """
    n = int(np.sqrt(len(x)))
    if n**2 != len(x):
        raise ValueError(
            "lenght of input array given as %s is not square of integer number!"
            % (len(x))
        )
    x_image = x.reshape(n, n)
    y_image = y.reshape(n, n)
    x_axes = x_image[0, :]
    y_axes = y_image[:, 0]
    return x_axes, y_axes


@export
def averaging(grid, numGrid, numPix):
    """Resize 2d pixel grid with numGrid to numPix and averages over the pixels.

    :param grid: higher resolution pixel grid
    :param numGrid: number of pixels per axis in the high resolution input image
    :param numPix: lower number of pixels per axis in the output image (numGrid/numPix
        is integer number)
    :return: averaged pixel grid
    """

    Nbig = numGrid
    Nsmall = numPix
    small = (
        grid.reshape([int(Nsmall), int(Nbig / Nsmall), int(Nsmall), int(Nbig / Nsmall)])
        .mean(3)
        .mean(1)
    )
    return small


@export
def displaceAbs(x, y, sourcePos_x, sourcePos_y):
    """Calculates a grid of distances to the observer in angel.

    :param x: cartesian coordinates
    :type x: numpy array
    :param y: cartesian coordinates
    :type y: numpy array
    :param sourcePos_x: source position
    :type sourcePos_x: float
    :param sourcePos_y: source position
    :type sourcePos_y: float
    :returns: array of displacement
    :raises: AttributeError, KeyError
    """
    x_mapped = x - sourcePos_x
    y_mapped = y - sourcePos_y
    absmapped = np.sqrt(x_mapped**2 + y_mapped**2)
    return absmapped


@export
def get_distance(x_mins, y_mins, x_true, y_true):
    """

    :param x_mins:
    :param y_mins:
    :param x_true:
    :param y_true:
    :return:
    """
    if len(x_mins) != len(x_true):
        return 10**10
    dist = 0
    x_true_list = np.array(x_true)
    y_true_list = np.array(y_true)

    for i in range(0, len(x_mins)):
        dist_list = (x_mins[i] - x_true_list) ** 2 + (y_mins[i] - y_true_list) ** 2
        dist += min(dist_list)
        k = np.where(dist_list == min(dist_list))
        if type(k) != int:
            k = k[0]
        x_true_list = np.delete(x_true_list, k)
        y_true_list = np.delete(y_true_list, k)
    return dist


@export
def compare_distance(x_mapped, y_mapped):
    """

    :param x_mapped: array of x-positions of remapped catalogue image
    :param y_mapped: array of y-positions of remapped catalogue image
    :return: sum of distance square of positions
    """
    X2 = 0
    for i in range(0, len(x_mapped) - 1):
        for j in range(i + 1, len(x_mapped)):
            dx = x_mapped[i] - x_mapped[j]
            dy = y_mapped[i] - y_mapped[j]
            X2 += dx**2 + dy**2
    return X2


@export
def min_square_dist(x_1, y_1, x_2, y_2):
    """Return minimum of quadratic distance of pairs (x1, y1) to pairs (x2, y2)

    :param x_1:
    :param y_1:
    :param x_2:
    :param y_2:
    :return:
    """
    dist = np.zeros_like(x_1)
    for i in range(len(x_1)):
        dist[i] = np.min((x_1[i] - x_2) ** 2 + (y_1[i] - y_2) ** 2)
    return dist


@export
def selectBest(array, criteria, numSelect, highest=True):
    """

    :param array: numpy array to be selected from
    :param criteria: criteria of selection
    :param highest: bool, if false the lowest will be selected
    :param numSelect: number of elements to be selected
    :return:
    """
    n = len(array)
    m = len(criteria)
    if n != m:
        raise ValueError(
            "Elements in array (%s) not equal to elements in criteria (%s)" % (n, m)
        )
    if n < numSelect:
        return array
    array_sorted = array[criteria.argsort()]
    if highest:
        result = array_sorted[n - numSelect :]
    else:
        result = array_sorted[0:numSelect]
    return result[::-1]


@export
def select_best(array, criteria, num_select, highest=True):
    """

    :param array: numpy array to be selected from
    :param criteria: criteria of selection
    :param highest: bool, if false the lowest will be selected
    :param num_select: number of elements to be selected
    :return:
    """
    n = len(array)
    m = len(criteria)
    if n != m:
        raise ValueError(
            "Elements in array (%s) not equal to elements in criteria (%s)" % (n, m)
        )
    if n < num_select:
        return array
    array = np.array(array)
    if highest is True:
        indexes = criteria.argsort()[::-1][:num_select]
    else:
        indexes = criteria.argsort()[::-1][n - num_select :]
    return array[indexes]


@export
def points_on_circle(radius, num_points, connect_ends=True):
    """Returns a set of uniform points around a circle.

    :param radius: radius of the circle
    :param num_points: number of points on the circle
    :param connect_ends: boolean, if True, start and end point are the same
    :return: x-coords, y-coords of points on the circle
    """
    if connect_ends:
        angle = np.linspace(0, 2 * np.pi, num_points)
    else:
        angle = np.linspace(0, 2 * np.pi * (1 - 1.0 / num_points), num_points)
    x_coord = np.cos(angle) * radius
    y_coord = np.sin(angle) * radius
    return x_coord, y_coord


@export
@jit()
def local_minima_2d(a, x, y):
    """Finds (local) minima in a 2d grid applies less rigid criteria for maximum without
    second-order tangential minima criteria.

    :param a: 1d array of displacements from the source positions
    :type a: numpy array with length numPix**2 in float
    :param x: 1d coordinate grid in x-direction
    :type x: numpy array with length numPix**2 in float
    :param y: 1d coordinate grid in x-direction
    :type y: numpy array with length numPix**2 in float
    :returns: array of indices of local minima, values of those minima
    :raises: AttributeError, KeyError
    """
    dim = int(np.sqrt(len(a)))
    values = []
    x_mins = []
    y_mins = []
    for i in range(dim + 1, len(a) - dim - 1):
        if (
            a[i] < a[i - 1]
            and a[i] < a[i + 1]
            and a[i] < a[i - dim]
            and a[i] < a[i + dim]
            and a[i] < a[i - (dim - 1)]
            and a[i] < a[i - (dim + 1)]
            and a[i] < a[i + (dim - 1)]
            and a[i] < a[i + (dim + 1)]
        ):
            x_mins.append(x[i])
            y_mins.append(y[i])
            values.append(a[i])
    return np.array(x_mins), np.array(y_mins), np.array(values)


@export
@jit()
def neighborSelect(a, x, y):
    """#TODO replace by from scipy.signal import argrelextrema for speed up >>> from
    scipy.signal import argrelextrema >>> x = np.array([2, 1, 2, 3, 2, 0, 1, 0]) >>>
    argrelextrema(x, np.greater) (array([3, 6]),) >>> y = np.array([[1, 2, 1, 2], ...
    [2, 2, 0, 0], ...               [5, 3, 4, 4]]) ... >>> argrelextrema(y, np.less,
    axis=1) (array([0, 2]), array([2, 1]))

    finds (local) minima in a 2d grid

    :param a: 1d array of displacements from the source positions
    :type a: numpy array with length numPix**2 in float
    :returns: array of indices of local minima, values of those minima
    :raises: AttributeError, KeyError
    """
    dim = int(np.sqrt(len(a)))
    values = []
    x_mins = []
    y_mins = []
    for i in range(dim + 1, len(a) - dim - 1):
        if (
            a[i] < a[i - 1]
            and a[i] < a[i + 1]
            and a[i] < a[i - dim]
            and a[i] < a[i + dim]
            and a[i] < a[i - (dim - 1)]
            and a[i] < a[i - (dim + 1)]
            and a[i] < a[i + (dim - 1)]
            and a[i] < a[i + (dim + 1)]
        ):
            if (
                a[i] < a[(i - 2 * dim - 1) % dim**2]
                and a[i] < a[(i - 2 * dim + 1) % dim**2]
                and a[i] < a[(i - dim - 2) % dim**2]
                and a[i] < a[(i - dim + 2) % dim**2]
                and a[i] < a[(i + dim - 2) % dim**2]
                and a[i] < a[(i + dim + 2) % dim**2]
                and a[i] < a[(i + 2 * dim - 1) % dim**2]
                and a[i] < a[(i + 2 * dim + 1) % dim**2]
                and a[i] < a[(i + 2 * dim) % dim**2]
                and a[i] < a[(i - 2 * dim) % dim**2]
                and a[i] < a[(i - 2) % dim**2]
                and a[i] < a[(i + 2) % dim**2]
            ):
                x_mins.append(x[i])
                y_mins.append(y[i])
                values.append(a[i])
    return np.array(x_mins), np.array(y_mins), np.array(values)


@export
def fwhm2sigma(fwhm):
    """

    :param fwhm: full-width-half-max value
    :return: gaussian sigma (sqrt(var))
    """
    sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
    return sigma


@export
def sigma2fwhm(sigma):
    """

    :param sigma:
    :return:
    """
    fwhm = sigma * (2 * np.sqrt(2 * np.log(2)))
    return fwhm


@export
def hyper2F2_array(a, b, c, d, x):
    """

    :param a:
    :param b:
    :param c:
    :param d:
    :param x:
    :return:
    """
    import mpmath

    if isinstance(x, int) or isinstance(x, float):
        out = mpmath.hyp2f2(a, b, c, d, x)
    else:
        n = len(x)
        out = np.zeros(n)
        for i in range(n):
            out[i] = mpmath.hyp2f2(a, b, c, d, x[i])
    return out


@export
def make_subgrid(ra_coord, dec_coord, subgrid_res=2):
    """Return a grid with subgrid resolution.

    :param ra_coord:
    :param dec_coord:
    :param subgrid_res:
    :return:
    """
    ra_array = array2image(ra_coord)
    dec_array = array2image(dec_coord)
    n = len(ra_array)
    d_ra_x = ra_array[0][1] - ra_array[0][0]
    d_ra_y = ra_array[1][0] - ra_array[0][0]
    d_dec_x = dec_array[0][1] - dec_array[0][0]
    d_dec_y = dec_array[1][0] - dec_array[0][0]

    ra_array_new = np.zeros((n * subgrid_res, n * subgrid_res))
    dec_array_new = np.zeros((n * subgrid_res, n * subgrid_res))
    for i in range(0, subgrid_res):
        for j in range(0, subgrid_res):
            ra_array_new[i::subgrid_res, j::subgrid_res] = (
                ra_array
                + d_ra_x * (-1 / 2.0 + 1 / (2.0 * subgrid_res) + j / float(subgrid_res))
                + d_ra_y * (-1 / 2.0 + 1 / (2.0 * subgrid_res) + i / float(subgrid_res))
            )
            dec_array_new[i::subgrid_res, j::subgrid_res] = (
                dec_array
                + d_dec_x
                * (-1 / 2.0 + 1 / (2.0 * subgrid_res) + j / float(subgrid_res))
                + d_dec_y
                * (-1 / 2.0 + 1 / (2.0 * subgrid_res) + i / float(subgrid_res))
            )

    ra_coords_sub = image2array(ra_array_new)
    dec_coords_sub = image2array(dec_array_new)
    return ra_coords_sub, dec_coords_sub


@export
def convert_bool_list(n, k=None):
    """Returns a bool list of the length of the lens models.

    if k = None: returns bool list with True's if k is int, returns bool list with
    False's but k'th is True if k is a list of int, e.g. [0, 3, 5], returns a bool list
    with True's in the integers listed and False elsewhere if k is a boolean list,
    checks for size to match the numbers of models and returns it

    :param n: integer, total length of output boolean list
    :param k: None, int, or list of ints
    :return: bool list
    """
    if k is None:
        bool_list = [True] * n
    elif isinstance(k, (int, np.integer)):  # single integer
        bool_list = [False] * n
        bool_list[k] = True
    elif len(k) == 0:  # empty list
        bool_list = [False] * n
    elif isinstance(k[0], bool):
        if n != len(k):
            raise ValueError(
                "length of selected lens models in format of boolean list is %s "
                "and does not match the models of this class instance %s." % (len(k), n)
            )
        bool_list = k
    elif isinstance(k[0], (int, np.integer)):  # list of integers
        bool_list = [False] * n
        for i, k_i in enumerate(k):
            if k_i is not False:
                # if k_i is True:
                #    bool_list[i] = True
                if k_i < n:
                    bool_list[k_i] = True
                else:
                    raise ValueError(
                        "k as set by %s is not convertable in a bool string of length %s !"
                        % (k, n)
                    )
    else:
        raise ValueError("input list k as %s not compatible" % k)
    return bool_list


def area(vs):
    """Use Green's theorem to compute the area enclosed by the given contour.

    param vs: 2d array of vertices of a contour line
    return: area within contour line
    """
    a = 0
    x0, y0 = vs[0]
    for [x1, y1] in vs[1:]:
        dx = x1 - x0
        dy = y1 - y0
        a += 0.5 * (y0 * dx - x0 * dy)
        x0 = x1
        y0 = y1
    return abs(a)



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.cnfw import CNFW
from lenstronomy.LensModel.Profiles.cnfw_ellipse import CNFW_ELLIPSE
import lenstronomy.Util.param_util as param_util

import numpy as np
import numpy.testing as npt
import pytest


class TestCNFWELLIPSE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.nfw = CNFW()
        self.nfw_e = CNFW_ELLIPSE()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        alpha_Rs = 1.0
        q = 1.0
        phi_G = 0
        r_core = 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.nfw.function(x, y, Rs, alpha_Rs, r_core=r_core)
        values_e = self.nfw_e.function(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(values[0], values_e[0], decimal=5)
        x = np.array([0])
        y = np.array([0])

        q = 0.8
        phi_G = 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.nfw_e.function(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(values[0], 0, decimal=4)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.nfw_e.function(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(values[0], 1.8550220596738973, decimal=5)
        npt.assert_almost_equal(values[1], 2.7684470762303537, decimal=5)
        npt.assert_almost_equal(values[2], 3.7076606717487586, decimal=5)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        alpha_Rs = 1.0
        q = 1.0
        phi_G = 0
        r_core = 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.nfw.derivatives(x, y, Rs, alpha_Rs, r_core)
        f_x_e, f_y_e = self.nfw_e.derivatives(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(f_x[0], f_x_e[0], decimal=5)
        npt.assert_almost_equal(f_y[0], f_y_e[0], decimal=5)
        x = np.array([0])
        y = np.array([0])
        alpha_Rs = 0
        f_x, f_y = self.nfw_e.derivatives(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(f_x[0], 0, decimal=5)
        npt.assert_almost_equal(f_y[0], 0, decimal=5)

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        alpha_Rs = 1.0
        q = 0.8
        phi_G = 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.nfw_e.derivatives(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(values[0][0], 0.3867896894988756, decimal=5)
        npt.assert_almost_equal(values[1][0], 1.1603690684966268, decimal=5)
        npt.assert_almost_equal(values[0][1], 0.9371571936062841, decimal=5)
        npt.assert_almost_equal(values[1][1], 0.46857859680314207, decimal=5)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        alpha_Rs = 1.0
        q = 1.0
        phi_G = 0
        r_core = 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.nfw.hessian(x, y, Rs, alpha_Rs, r_core)
        f_xx_e, f_xy_e, f_yx_e, f_yy_e = self.nfw_e.hessian(
            x, y, Rs, alpha_Rs, r_core, e1, e2
        )
        npt.assert_almost_equal(f_xx[0], f_xx_e[0], decimal=5)
        npt.assert_almost_equal(f_yy[0], f_yy_e[0], decimal=5)
        npt.assert_almost_equal(f_xy[0], f_xy_e[0], decimal=5)
        npt.assert_almost_equal(f_yx[0], f_yx_e[0], decimal=5)

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        q = 0.8
        phi_G = 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.nfw_e.hessian(x, y, Rs, alpha_Rs, r_core, e1, e2)
        npt.assert_almost_equal(values[0][0], 0.3306510620859626, decimal=5)
        npt.assert_almost_equal(values[3][0], 0.07493437759187316, decimal=5)
        npt.assert_almost_equal(values[1][0], -0.1684167189042185, decimal=5)
        npt.assert_almost_equal(values[0][1], 0.020280774837289073, decimal=5)
        npt.assert_almost_equal(values[3][1], 0.3955523575349673, decimal=5)
        npt.assert_almost_equal(values[1][1], -0.14605247788956888, decimal=5)

    def test_mass_3d(self):
        Rs = 10.0
        rho0 = 1.0
        r_core = 7.0

        R = np.linspace(0.1 * Rs, 4 * Rs, 1000)
        alpha_Rs = self.nfw._rho2alpha(rho0, Rs, r_core)
        m3d = self.nfw.mass_3d(R, Rs, rho0, r_core)
        m3d_lens = self.nfw_e.mass_3d_lens(R, Rs, alpha_Rs, r_core)
        npt.assert_almost_equal(m3d, m3d_lens, decimal=8)


if __name__ == "__main__":
    pytest.main()



"""
this file is taken from schwimmbad (https://github.com/adrn/schwimmbad) and an explicit fork by Aymeric Galan
to replace the multiprocessing with the multiprocess dependence as for multi-threading, multiprocessing is
not supporting dill (only pickle) which is required.

The class also extends with a ``is_master()`` definition

"""


# Standard library

import signal
import functools
import multiprocess
from multiprocess.pool import Pool

__all__ = ["MultiPool"]


def _initializer_wrapper(actual_initializer, *rest):
    """We ignore SIGINT.

    It's up to our parent to kill us in the typical condition of this arising from ``^C`` on a
    terminal. If someone is manually killing us with that signal, well... nothing will happen.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if actual_initializer is not None:
        actual_initializer(*rest)


class CallbackWrapper(object):
    def __init__(self, callback):
        self.callback = callback

    def __call__(self, tasks):
        for task in tasks:
            self.callback(task)


class MultiPool(Pool):
    """A modified version of :class:`multiprocessing.pool.Pool` that has better behavior
    with regard to ``KeyboardInterrupts`` in the :func:`map` method.

    (Original author: `Peter K. G. Williams <peter@newton.cx>`_)
    """

    wait_timeout = 3600

    def __init__(self, processes=None, initializer=None, initargs=(), **kwargs):
        """

        :param processes: The number of worker processes to use; defaults to the number of CPUs.
        :type processes: int, optional
        :param initializer: If specified, a callable that will be invoked by each worker process when it starts.
        :type initializer: callable, optional
        :param initargs: Arguments for ``initializer``; it will be called as ``initializer(*initargs)``.
        :type initargs: iterable, optional
        :param kwargs: Extra arguments passed to the :class:`multiprocessing.pool.Pool` superclass.
        """
        new_initializer = functools.partial(_initializer_wrapper, initializer)
        super(MultiPool, self).__init__(processes, new_initializer, initargs, **kwargs)
        self.size = self._processes
        self.rank = 0

    def is_master(self):
        return self.rank == 0

    def is_worker(self):
        return self.rank != 0

    @staticmethod
    def enabled():
        return True

    def map(self, func, iterable, chunksize=None, callback=None):
        """Equivalent to the built-in ``map()`` function and
        :meth:`multiprocessing.pool.Pool.map()`, without catching ``KeyboardInterrupt``.

        :param func: A function or callable object that is executed on each element of
            the specified ``tasks`` iterable. This object must be picklable
            (i.e. it can't be a function scoped within a function or a
            ``lambda`` function). This should accept a single positional
            argument and return a single object.
        :type func: callable
        :param iterable: A list or iterable of tasks. Each task can be itself an iterable
            (e.g., tuple) of values or data to pass in to the worker function.
        :type iterable: iterable
        :param callback: An optional callback function (or callable) that is called with the
            result from each worker run and is executed on the master process.
            This is useful for, e.g., saving results to a file, since the
            callback is only called on the master thread.
        :type callback: callable, optional

        :return: A list of results from the output of each ``worker()`` call.
        """

        if callback is None:
            callbackwrapper = None
        else:
            callbackwrapper = CallbackWrapper(callback)

        # The key magic is that we must call r.get() with a timeout, because
        # a Condition.wait() without a timeout swallows KeyboardInterrupts.
        r = self.map_async(
            func, iterable, chunksize=chunksize, callback=callbackwrapper
        )

        while True:
            try:
                return r.get(self.wait_timeout)

            except multiprocess.TimeoutError:
                pass

            except KeyboardInterrupt:
                self.terminate()
                self.join()
                raise



"""Provisional HST instrument and observational settings.

See Optics and Observation Conditions spreadsheet at
https://docs.google.com/spreadsheets/d/1pMUB_OOZWwXON2dd5oP8PekhCT5MBBZJO1HV7IMZg4Y/edit?usp=sharing
for list of
sources.
"""
import lenstronomy.Util.util as util

__all__ = ["HST"]

# F160W filter configs
WFC3_F160W_band_obs = {
    "exposure_time": 5400.0,  # ~90mins orbit on HST, but this number corresponds to
    # approximately two HST orbits with overheads, guide star aquisition. ~2700s science exposure per orbit
    "sky_brightness": 22.3,
    "magnitude_zero_point": 25.96,
    "num_exposures": 1,
    "seeing": 0.08,  # set equal to the approx pixel size for drizzled PSF. Note that undrizzled PSF FWHM ~ 0.15" (Windhorst et al 2011)
    "psf_type": "GAUSSIAN",
}

# configs meant to simulate images close to those provided as part of the Time Delay Lens Modeling Challenge
TDLMC_F160W_band_obs = {
    "exposure_time": 5400.0,  # ~90mins orbit on HST, but this number corresponds to
    # approximately two HST orbits with overheads, guide star aquisition. ~2700s science exposure per orbit
    "sky_brightness": 22.0,
    "magnitude_zero_point": 25.9463,
    "num_exposures": 1,
    "seeing": None,
    "psf_type": "PIXEL",  # note kernel_point_source (the PSF map) must be provided separately
}

# - keyword exposure_time: exposure time per image (in seconds)
# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword num_exposures: number of exposures that are combined (depends on coadd_years)
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' and 'PIXEL' supported)


class HST(object):
    """Class contains HST instrument and observation configurations."""

    def __init__(self, band="TDLMC_F160W", psf_type="PIXEL", coadd_years=None):
        """

        :param band: string, 'WFC3_F160W' or 'TDLMC_F160W' supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN', 'PIXEL' supported).
        :param coadd_years: int, number of years corresponding to num_exposures in obs dict. Currently supported: None.
        """

        if band == "TDLMC_F160W":
            self.obs = TDLMC_F160W_band_obs
        elif band == "WFC3_F160W" or band == "F160W":
            self.obs = WFC3_F160W_band_obs

        else:
            raise ValueError(
                "band %s not supported! Choose 'WFC3_F160W' or 'TDLMC_F160W'." % band
            )

        if psf_type == "GAUSSIAN":
            self.obs["psf_type"] = "GAUSSIAN"
        elif psf_type != "PIXEL":
            raise ValueError("psf_type %s not supported!" % psf_type)

        if coadd_years is not None:
            raise ValueError(
                " %s coadd_years not supported! "
                "You may manually adjust num_exposures in obs dict if required."
                % coadd_years
            )

        # WFC3 camera settings
        self.camera = {
            "read_noise": 4,
            "pixel_scale": 0.08,  # approx pixel size for drizzled PSF
            "ccd_gain": 2.5,
        }

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit)

    def kwargs_single_band(self):
        """

        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
import unittest
from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.Solver.solver2point import Solver2Point
import lenstronomy.Util.param_util as param_util


class TestSolver(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        pass

    def test_subtract(self):
        lensModel = LensModel(["SPEP"])
        solver_spep_center = Solver2Point(lensModel, solver_type="CENTER")
        x_cat = np.array([0, 0])
        y_cat = np.array([1, 2])
        a = solver_spep_center._subtract_constraint(x_cat, y_cat)
        assert a[0] == 0
        assert a[1] == 1

    def test_all_spep(self):
        lensModel = LensModel(["SPEP"])
        solver_spep_center = Solver2Point(lensModel, solver_type="CENTER")
        solver_spep_ellipse = Solver2Point(lensModel, solver_type="ELLIPSE")
        image_position_spep = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        gamma = 1.9
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": 1,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            }
        ]
        x_pos, y_pos = image_position_spep.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=0.01,
            search_window=5,
            precision_limit=10 ** (-10),
            num_iter_max=10,
        )
        print(x_pos, y_pos, "test")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]
        kwargs_init = [
            {
                "theta_E": 1,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        kwargs_out_center, precision = solver_spep_center.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )

        kwargs_init = [
            {
                "theta_E": 1,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.1,
                "center_y": -0.1,
            }
        ]
        kwargs_out_ellipse, precision = solver_spep_ellipse.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )

        npt.assert_almost_equal(
            kwargs_out_center[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out_center[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )
        npt.assert_almost_equal(kwargs_out_center[0]["center_y"], -0.1, decimal=3)

        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(kwargs_out_ellipse[0]["e1"], e1, decimal=3)

    def test_all_nfw(self):
        lensModel = LensModel(["SPEP"])
        solver_nfw_ellipse = Solver2Point(lensModel, solver_type="ELLIPSE")
        solver_nfw_center = Solver2Point(lensModel, solver_type="CENTER")
        spep = LensModel(["SPEP"])

        image_position_nfw = LensEquationSolver(LensModel(["SPEP", "NFW"]))
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        deltapix = 0.05
        numPix = 100
        gamma = 1.9
        Rs = 0.1
        nfw = NFW()
        alpha_Rs = nfw.rho02alpha(1.0, Rs)
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"Rs": Rs, "alpha_Rs": alpha_Rs, "center_x": -0.5, "center_y": 0.5},
        ]
        x_pos, y_pos = image_position_nfw.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(len(x_pos), "number of images")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]

        kwargs_init = [
            {
                "theta_E": 1,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"Rs": Rs, "alpha_Rs": alpha_Rs, "center_x": -0.5, "center_y": 0.5},
        ]
        kwargs_out_center, precision = solver_nfw_center.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )
        source_x, source_y = spep.ray_shooting(x_pos[0], y_pos[0], kwargs_out_center)
        x_pos_new, y_pos_new = image_position_nfw.findBrightImage(
            source_x,
            source_y,
            kwargs_out_center,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        print(kwargs_out_center, "kwargs_out_center")
        npt.assert_almost_equal(x_pos_new[0], x_pos[0], decimal=2)
        npt.assert_almost_equal(y_pos_new[0], y_pos[0], decimal=2)

        npt.assert_almost_equal(
            kwargs_out_center[0]["center_x"], kwargs_lens[0]["center_x"], decimal=2
        )
        npt.assert_almost_equal(
            kwargs_out_center[0]["center_y"], kwargs_lens[0]["center_y"], decimal=2
        )
        npt.assert_almost_equal(kwargs_out_center[0]["center_y"], -0.1, decimal=2)

        kwargs_init = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"Rs": Rs, "alpha_Rs": alpha_Rs, "center_x": -0.5, "center_y": 0.5},
        ]
        kwargs_out_ellipse, precision = solver_nfw_ellipse.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )

        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e1"], kwargs_lens[0]["e1"], decimal=2
        )
        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e2"], kwargs_lens[0]["e2"], decimal=2
        )
        npt.assert_almost_equal(kwargs_out_ellipse[0]["e1"], e1, decimal=2)

    def test_all_spep_sis(self):
        lensModel = LensModel(["SPEP", "SIS"])
        solver_ellipse = Solver2Point(lensModel, solver_type="ELLIPSE")
        solver_center = Solver2Point(lensModel, solver_type="CENTER")
        spep = LensModel(["SPEP", "SIS"])
        image_position = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        deltapix = 0.05
        numPix = 100
        gamma = 1.9
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.6, "center_x": -0.5, "center_y": 0.5},
        ]
        x_pos, y_pos = image_position.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
            precision_limit=10 ** (-10),
        )
        print(len(x_pos), "number of images")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]

        kwargs_init = [
            {
                "theta_E": 1,
                "gamma": gamma,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"theta_E": 0.6, "center_x": -0.5, "center_y": 0.5},
        ]
        kwargs_out_center, precision = solver_center.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )
        print(kwargs_out_center, "output")
        source_x, source_y = spep.ray_shooting(x_pos[0], y_pos[0], kwargs_out_center)
        x_pos_new, y_pos_new = image_position.findBrightImage(
            source_x,
            source_y,
            kwargs_out_center,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        npt.assert_almost_equal(x_pos_new[0], x_pos[0], decimal=3)
        npt.assert_almost_equal(y_pos_new[0], y_pos[0], decimal=3)

        npt.assert_almost_equal(
            kwargs_out_center[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out_center[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )
        npt.assert_almost_equal(kwargs_out_center[0]["center_y"], -0.1, decimal=3)

        kwargs_init = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.6, "center_x": -0.5, "center_y": 0.5},
        ]
        kwargs_out_ellipse, precision = solver_ellipse.constraint_lensmodel(
            x_pos, y_pos, kwargs_init
        )

        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out_ellipse[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(kwargs_out_ellipse[0]["e1"], 0.2, decimal=3)

    def test_shapelet_cart(self):
        lens_model_list = ["SHAPELETS_CART", "SIS"]
        lens = LensModel(lens_model_list)
        solver = Solver2Point(lens, solver_type="SHAPELETS")
        image_position = LensEquationSolver(lens)
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        deltapix = 0.05
        numPix = 100

        kwargs_lens = [
            {"coeffs": [1.0, 0.0, 0.1, 1.0], "beta": 1.0},
            {"theta_E": 1.0, "center_x": -0.1, "center_y": 0.1},
        ]
        x_pos, y_pos = image_position.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
            precision_limit=10 ** (-10),
        )
        print(len(x_pos), "number of images")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]

        kwargs_init = [
            {"coeffs": [1.0, 0.0, 0.1, 1.0], "beta": 1.0},
            {"theta_E": 1.0, "center_x": -0.1, "center_y": 0.1},
        ]
        kwargs_out, precision = solver.constraint_lensmodel(x_pos, y_pos, kwargs_init)
        print(kwargs_out, "output")
        source_x, source_y = lens.ray_shooting(x_pos[0], y_pos[0], kwargs_out)
        x_pos_new, y_pos_new = image_position.findBrightImage(
            source_x,
            source_y,
            kwargs_out,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        npt.assert_almost_equal(x_pos_new[0], x_pos[0], decimal=3)
        npt.assert_almost_equal(y_pos_new[0], y_pos[0], decimal=3)

        npt.assert_almost_equal(
            kwargs_out[0]["coeffs"][1], kwargs_lens[0]["coeffs"][1], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out[0]["coeffs"][2], kwargs_lens[0]["coeffs"][2], decimal=3
        )

    def test_theta_E_phi(self):
        lensModel = LensModel(["SPEP", "SHEAR"])
        solver = Solver2Point(lensModel, solver_type="THETA_E_PHI")

        image_position = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        deltapix = 0.05
        numPix = 100
        gamma = 1.9
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.1,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma1": 0.03, "gamma2": 0.0},
        ]
        x_pos, y_pos = image_position.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
            precision_limit=10 ** (-15),
        )
        print(len(x_pos), "number of images")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]

        kwargs_init = [
            {
                "theta_E": 1.9,
                "gamma": gamma,
                "e1": 0.1,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma1": 0.0, "gamma2": 0.03},
        ]
        kwargs_out, precision = solver.constraint_lensmodel(x_pos, y_pos, kwargs_init)
        print(kwargs_out, "output")
        source_x, source_y = lensModel.ray_shooting(x_pos[0], y_pos[0], kwargs_out)
        x_pos_new, y_pos_new = image_position.findBrightImage(
            source_x,
            source_y,
            kwargs_out,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        npt.assert_almost_equal(x_pos_new[0], x_pos[0], decimal=3)
        npt.assert_almost_equal(y_pos_new[0], y_pos[0], decimal=3)

        npt.assert_almost_equal(
            kwargs_out[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_out[1]["gamma1"], kwargs_lens[1]["gamma1"], decimal=2
        )
        npt.assert_almost_equal(
            kwargs_out[1]["gamma2"], kwargs_lens[1]["gamma2"], decimal=2
        )

    def test_theta_E_ellipse(self):
        lensModel = LensModel(["SPEP", "SHEAR"])
        solver = Solver2Point(lensModel, solver_type="THETA_E_ELLIPSE")

        image_position = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = 0.03
        deltapix = 0.05
        numPix = 100
        gamma = 1.9
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.1,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma1": 0.03, "gamma2": 0.0},
        ]
        x_pos, y_pos = image_position.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
            precision_limit=10 ** (-15),
        )
        print(len(x_pos), "number of images")
        x_pos = x_pos[:2]
        y_pos = y_pos[:2]

        kwargs_init = [
            {
                "theta_E": 1.9,
                "gamma": gamma,
                "e1": -0.03,
                "e2": 0.1,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma1": 0.03, "gamma2": 0.0},
        ]
        kwargs_out, precision = solver.constraint_lensmodel(x_pos, y_pos, kwargs_init)
        print(kwargs_out, "output")
        source_x, source_y = lensModel.ray_shooting(x_pos[0], y_pos[0], kwargs_out)
        x_pos_new, y_pos_new = image_position.findBrightImage(
            source_x,
            source_y,
            kwargs_out,
            numImages=2,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        npt.assert_almost_equal(x_pos_new[0], x_pos[0], decimal=3)
        npt.assert_almost_equal(y_pos_new[0], y_pos[0], decimal=3)

        npt.assert_almost_equal(
            kwargs_out[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(kwargs_out[0]["e1"], kwargs_lens[0]["e1"], decimal=2)
        npt.assert_almost_equal(kwargs_out[0]["e2"], kwargs_lens[0]["e2"], decimal=2)

    def test_add_fixed_lens(self):
        lensModel = LensModel(lens_model_list=["SPEP", "SHEAR"])
        kwargs_lens_init = [{"theta_E": 1, "e2": 0}, {"gamma2": 0}]
        solver = Solver2Point(
            lensModel=lensModel, solver_type="THETA_E_PHI", decoupling=True
        )
        kwargs_fixed_added = solver.add_fixed_lens(
            kwargs_fixed_lens_list=[{}, {}], kwargs_lens_init=kwargs_lens_init
        )
        print(kwargs_fixed_added)
        assert kwargs_fixed_added[1]["gamma2"] == 0

        solver = Solver2Point(
            lensModel=lensModel, solver_type="THETA_E_ELLIPSE", decoupling=True
        )
        kwargs_fixed_added = solver.add_fixed_lens(
            kwargs_fixed_lens_list=[{}, {}], kwargs_lens_init=kwargs_lens_init
        )
        assert kwargs_fixed_added[0]["e2"] == 0

        lensModel = LensModel(lens_model_list=["SHAPELETS_CART"])
        solver = Solver2Point(
            lensModel=lensModel, solver_type="SHAPELETS", decoupling=True
        )
        kwargs_fixed_added = solver.add_fixed_lens(
            kwargs_fixed_lens_list=[{}, {}], kwargs_lens_init=kwargs_lens_init
        )
        assert len(kwargs_fixed_added) == len(kwargs_lens_init)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP"])
            Solver2Point(lensModel=lensModel, solver_type="WRONG", decoupling=True)
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP"])
            Solver2Point(lensModel=lensModel, solver_type="SHAPELETS", decoupling=True)
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP", "SIS"])
            Solver2Point(
                lensModel=lensModel, solver_type="THETA_E_PHI", decoupling=True
            )
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP", "SHEAR"])
            solver = Solver2Point(
                lensModel=lensModel, solver_type="THETA_E_PHI", decoupling=True
            )
            solver._solver_type = "WRONG"
            solver._update_kwargs(x=None, kwargs_list=None)
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP", "SHEAR"])
            solver = Solver2Point(
                lensModel=lensModel, solver_type="THETA_E_PHI", decoupling=True
            )
            solver._solver_type = "WRONG"
            solver._extract_array(kwargs_list=None)
        with self.assertRaises(ValueError):
            lensModel = LensModel(lens_model_list=["SPEP", "SHEAR"])
            solver = Solver2Point(
                lensModel=lensModel, solver_type="THETA_E_PHI", decoupling=True
            )
            solver._solver_type = "WRONG"
            solver.add_fixed_lens(
                kwargs_fixed_lens_list=[None], kwargs_lens_init=[None]
            )


if __name__ == "__main__":
    pytest.main()



import lenstronomy.Util.kin_sampling_util as kin_sampling_util

import numpy as np
import pytest
import numpy.testing as npt
import matplotlib.pyplot as plt


class TestKinSamplingUtil(object):
    def setup(self):
        self.image = np.array(
            [
                [0, 1, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )

        transform_pix2angle = np.array(
            [[1, 0], [0, 1]]
        )  # linear translation matrix of no rotation
        ra_at_xy_0, dec_at_xy_0 = transform_pix2angle.dot([-5 / 2, -5 / 2]).T
        self.kwargs_pixel_rot_imaging = {
            "nx": 5,
            "ny": 5,  # number of pixels per axis
            "ra_at_xy_0": ra_at_xy_0,  # RA at pixel (0,0)
            "dec_at_xy_0": dec_at_xy_0,  # DEC at pixel (0,0)
            "transform_pix2angle": transform_pix2angle,
        }
        self.imaging_inputs = {
            "image": self.image,
            "transform_pix2angle": self.kwargs_pixel_rot_imaging["transform_pix2angle"],
            "ra_at_xy0": self.kwargs_pixel_rot_imaging["ra_at_xy_0"],
            "dec_at_xy0": self.kwargs_pixel_rot_imaging["dec_at_xy_0"],
            "ellipse_PA": 0,
            "offset_x": 0,
            "offset_y": 0,
        }
        self.kinNN_inputs = {"image": self.image, "deltaPix": 1}

    def define_spectra_kwargs_from_rotation_angle(self, rotation_angle):
        transform_pix2angle = np.array(
            [
                [np.cos(rotation_angle), -np.sin(rotation_angle)],
                [np.sin(rotation_angle), np.cos(rotation_angle)],
            ]
        )  # linear translation matrix of a 90 degree rotation
        ra_at_xy_0, dec_at_xy_0 = transform_pix2angle.dot([-5 / 2, -5 / 2]).T
        kwargs_pixel_rot_spectra = {
            "nx": 5,
            "ny": 5,  # number of pixels per axis
            "ra_at_xy_0": ra_at_xy_0,  # RA at pixel (0,0)
            "dec_at_xy_0": dec_at_xy_0,  # DEC at pixel (0,0)
            "transform_pix2angle": transform_pix2angle,
        }
        spectra_inputs = {
            "image": self.image,
            "transform_pix2angle": kwargs_pixel_rot_spectra["transform_pix2angle"],
            "ra_at_xy0": kwargs_pixel_rot_spectra["ra_at_xy_0"],
            "dec_at_xy0": kwargs_pixel_rot_spectra["dec_at_xy_0"],
        }
        return spectra_inputs

    def test_rotation(self):
        # spectra grid not rotated from imaging:
        spectra_inputs = self.define_spectra_kwargs_from_rotation_angle(0)
        skinn_align = kin_sampling_util.KinNNImageAlign(
            spectra_inputs, self.imaging_inputs, self.kinNN_inputs
        )
        spectra_img_coords = skinn_align.spectragrid_in_imagingxy()
        imaging_img_coords = skinn_align.pix_coords(self.imaging_inputs)
        npt.assert_allclose(imaging_img_coords, spectra_img_coords, atol=1e-7)

        spectra_inputs = self.define_spectra_kwargs_from_rotation_angle(2 * np.pi)
        skinn_align = kin_sampling_util.KinNNImageAlign(
            spectra_inputs, self.imaging_inputs, self.kinNN_inputs
        )
        spectra_img_coords = skinn_align.spectragrid_in_imagingxy()
        imaging_img_coords = skinn_align.pix_coords(self.imaging_inputs)
        npt.assert_allclose(imaging_img_coords, spectra_img_coords, atol=1e-7)

        # spectra_grid 90 degrees from imaging:
        spectra_inputs = self.define_spectra_kwargs_from_rotation_angle(np.pi / 2)
        skinn_align = kin_sampling_util.KinNNImageAlign(
            spectra_inputs, self.imaging_inputs, self.kinNN_inputs
        )
        spectra_img_coords = skinn_align.spectragrid_in_imagingxy()
        imaging_img_coords = skinn_align.pix_coords(self.imaging_inputs)
        npt.assert_allclose(
            imaging_img_coords[0], spectra_img_coords[1], atol=1e-7
        )  # old x coordinate is new y coordinate

    def test_samegrid(self):
        # check interpolation in case where grids are the same
        spectra_inputs = self.define_spectra_kwargs_from_rotation_angle(0)
        skinn_align = kin_sampling_util.KinNNImageAlign(
            spectra_inputs, self.imaging_inputs, self.kinNN_inputs
        )
        interp_image = skinn_align.interp_image()
        npt.assert_almost_equal(interp_image, self.image)

    def test_offset(self):
        imaging_inputs = {
            "image": self.image,
            "transform_pix2angle": self.kwargs_pixel_rot_imaging["transform_pix2angle"],
            "ra_at_xy0": self.kwargs_pixel_rot_imaging["ra_at_xy_0"],
            "dec_at_xy0": self.kwargs_pixel_rot_imaging["dec_at_xy_0"],
            "ellipse_PA": 0,
            "offset_x": 1,
            "offset_y": 1,
        }
        spectra_inputs = self.define_spectra_kwargs_from_rotation_angle(0)
        skinn_align = kin_sampling_util.KinNNImageAlign(
            spectra_inputs, imaging_inputs, self.kinNN_inputs
        )
        interp_image = skinn_align.interp_image()
        # #uncomment to see plots
        # plt.imshow(image)
        # plt.show()
        # plt.imshow(interp_image)
        # plt.show()
        expected_img = np.array(
            [
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
        npt.assert_allclose(interp_image, expected_img, atol=1e-7)



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
from lenstronomy.ImSim.Numerics.numba_convolution import (
    NumbaConvolution,
    SubgridNumbaConvolution,
)
from lenstronomy.ImSim.Numerics.convolution import PixelKernelConvolution
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Util import util
from lenstronomy.Util import image_util
import pytest


class TestPixelConvolution(object):
    def setup_method(self):
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1
        self.num_pix = 10
        self.num_pix_kernel = 7
        x, y = util.make_grid(numPix=self.num_pix_kernel, deltapix=self.delta_pix)
        kwargs_kernel = [{"amp": 1, "sigma": 3, "center_x": 0, "center_y": 0}]
        kernel = lightModel.surface_brightness(x, y, kwargs_kernel)
        self.kernel = util.array2image(kernel)
        self.kernel /= np.sum(self.kernel)

        x, y = util.make_grid(numPix=self.num_pix, deltapix=self.delta_pix)
        kwargs = [{"amp": 1, "sigma": 2, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)

    def test_convolve2d(self):
        conv_pixels = np.ones_like(self.model)
        conv_pixels = np.array(conv_pixels, dtype=bool)
        numba_conv = NumbaConvolution(
            kernel=self.kernel, conv_pixels=conv_pixels, compute_pixels=conv_pixels
        )

        model_conv_numba = numba_conv.convolve2d(self.model)

        pixel_conv = PixelKernelConvolution(kernel=self.kernel)
        image_convolved = pixel_conv.convolution2d(self.model)
        npt.assert_almost_equal(model_conv_numba, image_convolved, decimal=10)


class TestSubgirdNumbaConvolution(object):
    def setup_method(self):
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.supersampling_factor = 3
        self.delta_pix = 1
        self.num_pix = 10
        self.num_pix_kernel = 7
        x, y = util.make_grid(numPix=self.num_pix_kernel, deltapix=self.delta_pix)
        kwargs_kernel = [{"amp": 1, "sigma": 3, "center_x": 0, "center_y": 0}]
        kernel = lightModel.surface_brightness(x, y, kwargs_kernel)
        self.kernel = util.array2image(kernel)
        self.kernel /= np.sum(self.kernel)

        x_sub, y_sub = util.make_grid(
            numPix=self.num_pix_kernel,
            deltapix=self.delta_pix,
            subgrid_res=self.supersampling_factor,
        )
        kernel_super = lightModel.surface_brightness(x_sub, y_sub, kwargs_kernel)
        self.kernel_super = util.array2image(kernel_super)
        self.kernel_super /= np.sum(self.kernel_super)

        x_sub, y_sub = util.make_grid(
            numPix=self.num_pix,
            deltapix=self.delta_pix,
            subgrid_res=self.supersampling_factor,
        )
        kwargs = [{"amp": 1, "sigma": 2, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x_sub, y_sub, kwargs)
        self.model_super = util.array2image(flux)
        self.model = image_util.re_size(
            self.model_super, factor=self.supersampling_factor
        )

    def test_convolve2d(self):
        conv_pixels = np.ones_like(self.model)
        conv_pixels = np.array(conv_pixels, dtype=bool)
        numba_conv = SubgridNumbaConvolution(
            kernel_super=self.kernel_super,
            conv_pixels=conv_pixels,
            compute_pixels=conv_pixels,
            supersampling_factor=self.supersampling_factor,
        )

        model_conv_numba = numba_conv.convolve2d(self.model_super)
        pixel_conv = PixelKernelConvolution(kernel=self.kernel_super)
        image_convolved = pixel_conv.convolution2d(self.model_super)
        image_convolved = image_util.re_size(
            image_convolved, factor=self.supersampling_factor
        )
        npt.assert_almost_equal(model_conv_numba, image_convolved, decimal=10)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
from scipy import ndimage
from scipy import interpolate
import copy
import lenstronomy.Util.util as util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def add_layer2image(grid2d, x_pos, y_pos, kernel, order=1):
    """Adds a kernel on the grid2d image at position x_pos, y_pos with an interpolated
    subgrid pixel shift of order=order.

    :param grid2d: 2d pixel grid (i.e. image)
    :param x_pos: x-position center (pixel coordinate) of the layer to be added
    :param y_pos: y-position center (pixel coordinate) of the layer to be added
    :param kernel: the layer to be added to the image
    :param order: interpolation order for sub-pixel shift of the kernel to be added
    :return: image with added layer, cut to original size
    """

    x_int = int(round(x_pos))
    y_int = int(round(y_pos))
    shift_x = x_int - x_pos
    shift_y = y_int - y_pos
    kernel_shifted = ndimage.shift(kernel, shift=[-shift_y, -shift_x], order=order)
    return add_layer2image_int(grid2d, x_int, y_int, kernel_shifted)


@export
def add_layer2image_int(grid2d, x_pos, y_pos, kernel):
    """Adds a kernel on the grid2d image at position x_pos, y_pos at integer positions
    of pixel.

    :param grid2d: 2d pixel grid (i.e. image)
    :param x_pos: x-position center (pixel coordinate) of the layer to be added
    :param y_pos: y-position center (pixel coordinate) of the layer to be added
    :param kernel: the layer to be added to the image
    :return: image with added layer
    """
    nx, ny = np.shape(kernel)
    if nx % 2 == 0:
        raise ValueError("kernel needs odd numbers of pixels")

    num_x, num_y = np.shape(grid2d)
    x_int = int(round(x_pos))
    y_int = int(round(y_pos))

    k_x, k_y = np.shape(kernel)
    k_l2_x = int((k_x - 1) / 2)
    k_l2_y = int((k_y - 1) / 2)

    min_x = np.maximum(0, x_int - k_l2_x)
    min_y = np.maximum(0, y_int - k_l2_y)
    max_x = np.minimum(num_x, x_int + k_l2_x + 1)
    max_y = np.minimum(num_y, y_int + k_l2_y + 1)

    min_xk = np.maximum(0, -x_int + k_l2_x)
    min_yk = np.maximum(0, -y_int + k_l2_y)
    max_xk = np.minimum(k_x, -x_int + k_l2_x + num_x)
    max_yk = np.minimum(k_y, -y_int + k_l2_y + num_y)
    if (
        min_x >= max_x
        or min_y >= max_y
        or min_xk >= max_xk
        or min_yk >= max_yk
        or (max_x - min_x != max_xk - min_xk)
        or (max_y - min_y != max_yk - min_yk)
    ):
        return grid2d
    kernel_re_sized = kernel[min_yk:max_yk, min_xk:max_xk]
    new = grid2d.copy()
    new[min_y:max_y, min_x:max_x] += kernel_re_sized
    return new


@export
def add_background(image, sigma_bkd):
    """Generates background noise to image. To generate a noisy image with background
    noise, generate image_noisy = image + add_background(image, sigma_bkd)

    :param image: pixel values of image
    :param sigma_bkd: background noise (sigma)
    :return: a realisation of Gaussian noise of the same size as image
    """
    nx, ny = np.shape(image)
    background = np.random.randn(nx, ny) * sigma_bkd
    return background


@export
def add_poisson(image, exp_time):
    """Generates a poison (or Gaussian) distributed noise with mean given by surface
    brightness. To generate a noisy image with Poisson noise, perform image_noisy =
    image + add_poisson(image, exp_time)

    :param image: pixel values (photon counts per unit exposure time)
    :param exp_time: exposure time
    :return: Poisson noise realization of input image
    """

    # Gaussian approximation for Poisson distribution, normalized to exposure time
    sigma = np.sqrt(np.abs(image) / exp_time)
    nx, ny = np.shape(image)
    poisson = np.random.randn(nx, ny) * sigma
    return poisson


@export
def rotateImage(img, angle):
    """Querries scipy.ndimage.rotate routine :param img: image to be rotated :param
    angle: angle to be rotated (radian) :return: rotated image."""
    imgR = ndimage.rotate(img, angle, reshape=False)
    return imgR


@export
def shift_image(img, shift):
    """Queries scipy.ndimage.shift routine.

    :param img: image to be shifted
    :param shift: sequence containing x and y shift in pixels
    :return: shifted image
    """
    img_s = ndimage.shift(img, shift)
    return img_s


@export
def re_size_array(x_in, y_in, input_values, x_out, y_out):
    """Resizes 2d array (i.e. image) to new coordinates. So far only works with square
    output aligned with coordinate axis.

    :param x_in:
    :param y_in:
    :param input_values:
    :param x_out:
    :param y_out:
    :return:
    """
    # from skimage.transform import resize
    # resize(input_values)
    interp_2d = interpolate.interp2d(x_in, y_in, input_values, kind="linear")
    # interp_2d = scipy.interpolate.RectBivariateSpline(x_in, y_in, input_values, kx=1, ky=1)
    out_values = interp_2d.__call__(x_out, y_out)
    return out_values


@export
def symmetry_average(image, symmetry):
    """Symmetry averaged image.

    :param image:
    :param symmetry:
    :return:
    """
    img_sym = np.zeros_like(image)
    angle = 360.0 / symmetry
    for i in range(symmetry):
        img_sym += rotateImage(image, angle * i)
    img_sym /= symmetry
    return img_sym


@export
def findOverlap(x_mins, y_mins, min_distance):
    """Finds overlapping solutions, deletes multiples and deletes non-solutions and if
    it is not a solution, deleted as well."""
    n = len(x_mins)
    idex = []
    for i in range(n):
        if i == 0:
            pass
        else:
            for j in range(0, i):
                if (
                    abs(x_mins[i] - x_mins[j]) < min_distance
                    and abs(y_mins[i] - y_mins[j]) < min_distance
                ):
                    idex.append(i)
                    break
    x_mins = np.delete(x_mins, idex, axis=0)
    y_mins = np.delete(y_mins, idex, axis=0)
    return x_mins, y_mins


@export
def coordInImage(x_coord, y_coord, num_pix, deltapix):
    """
    checks whether image positions are within the pixel image in units of arcsec
    if not: remove it

    :returns: image positions within the pixel image
    """
    idex = []
    min_ = -deltapix * num_pix / 2
    max_ = deltapix * num_pix / 2
    for i in range(len(x_coord)):  # sum over image positions
        if (
            x_coord[i] < min_
            or x_coord[i] > max_
            or y_coord[i] < min_
            or y_coord[i] > max_
        ):
            idex.append(i)
    x_coord = np.delete(x_coord, idex, axis=0)
    y_coord = np.delete(y_coord, idex, axis=0)
    return x_coord, y_coord


@export
def re_size(image, factor=1):
    """Re-sizes image with nx x ny to nx/factor x ny/factor.

    :param image: 2d image with shape (nx,ny)
    :param factor: integer >=1
    :return:
    """
    if factor < 1:
        raise ValueError("scaling factor in re-sizing %s < 1" % factor)
    elif factor == 1:
        return image
    f = int(factor)
    nx, ny = np.shape(image)
    if int(nx / f) == nx / f and int(ny / f) == ny / f:
        small = image.reshape([int(nx / f), f, int(ny / f), f]).mean(3).mean(1)
        return small
    else:
        raise ValueError(
            "scaling with factor %s is not possible with grid size %s, %s" % (f, nx, ny)
        )


@export
def rebin_image(bin_size, image, wht_map, sigma_bkg, ra_coords, dec_coords, idex_mask):
    """Re-bins pixels, updates cutout image, wht_map, sigma_bkg, coordinates, PSF.

    :param bin_size: number of pixels (per axis) to merge
    :return:
    """
    numPix = int(len(image) / bin_size)
    numPix_precut = numPix * bin_size
    factor = int(len(image) / numPix)
    if not numPix * bin_size == len(image):
        image_precut = image[0:numPix_precut, 0:numPix_precut]
    else:
        image_precut = image
    image_resized = re_size(image_precut, factor)
    image_resized *= bin_size**2
    wht_map_resized = re_size(wht_map[0:numPix_precut, 0:numPix_precut], factor)
    sigma_bkg_resized = bin_size * sigma_bkg
    ra_coords_resized = re_size(ra_coords[0:numPix_precut, 0:numPix_precut], factor)
    dec_coords_resized = re_size(dec_coords[0:numPix_precut, 0:numPix_precut], factor)
    idex_mask_resized = re_size(idex_mask[0:numPix_precut, 0:numPix_precut], factor)
    idex_mask_resized[idex_mask_resized > 0] = 1
    return (
        image_resized,
        wht_map_resized,
        sigma_bkg_resized,
        ra_coords_resized,
        dec_coords_resized,
        idex_mask_resized,
    )


@export
def rebin_coord_transform(factor, x_at_radec_0, y_at_radec_0, Mpix2coord, Mcoord2pix):
    """Adopt coordinate system and transformation between angular and pixel coordinates
    of a re-binned image."""
    factor = int(factor)
    Mcoord2pix_resized = Mcoord2pix / factor
    Mpix2coord_resized = Mpix2coord * factor
    x_at_radec_0_resized = (x_at_radec_0 + 0.5) / factor - 0.5
    y_at_radec_0_resized = (y_at_radec_0 + 0.5) / factor - 0.5
    ra_at_xy_0_resized, dec_at_xy_0_resized = util.map_coord2pix(
        -x_at_radec_0_resized, -y_at_radec_0_resized, 0, 0, Mpix2coord_resized
    )
    return (
        ra_at_xy_0_resized,
        dec_at_xy_0_resized,
        x_at_radec_0_resized,
        y_at_radec_0_resized,
        Mpix2coord_resized,
        Mcoord2pix_resized,
    )


@export
def stack_images(image_list, wht_list, sigma_list):
    """Stacks images and saves new image as a fits file.

    :return:
    """
    image_stacked = np.zeros_like(image_list[0])
    wht_stacked = np.zeros_like(image_stacked)
    sigma_stacked = 0.0
    for i in range(len(image_list)):
        image_stacked += image_list[i] * wht_list[i]
        sigma_stacked += sigma_list[i] ** 2 * np.median(wht_list[i])
        wht_stacked += wht_list[i]
    image_stacked /= wht_stacked
    sigma_stacked /= np.median(wht_stacked)
    wht_stacked /= len(wht_list)
    return image_stacked, wht_stacked, np.sqrt(sigma_stacked)


@export
def cut_edges(image, num_pix):
    """Cuts out the edges of a 2d image and returns re-sized image to numPix center is
    well defined for odd pixel sizes.

    :param image: 2d numpy array
    :param num_pix: square size of cut out image
    :return: cutout image with size numPix
    """
    nx, ny = image.shape
    if nx < num_pix or ny < num_pix:
        raise ValueError(
            "image can not be resized, in routine cut_edges with image shape (%s %s) "
            "and desired new shape (%s %s)" % (nx, ny, num_pix, num_pix)
        )
    if (nx % 2 == 0 and ny % 2 == 1) or (nx % 2 == 1 and ny % 2 == 0):
        raise ValueError(
            "image with odd and even axis (%s %s) not supported for re-sizing"
            % (nx, ny)
        )
    if (nx % 2 == 0 and num_pix % 2 == 1) or (nx % 2 == 1 and num_pix % 2 == 0):
        raise ValueError(
            "image can only be re-sized from even to even or odd to odd number."
        )

    x_min = int((nx - num_pix) / 2)
    y_min = int((ny - num_pix) / 2)
    x_max = nx - x_min
    y_max = ny - y_min
    resized = image[x_min:x_max, y_min:y_max]
    return copy.deepcopy(resized)


@export
def radial_profile(data, center):
    """Computes radial profile.

    :param data: 2d numpy array
    :param center: center [x, y] from which pixel to compute the radial profile
    :return: radial profile (in units pixel)
    """
    y, x = np.indices(data.shape)
    r = np.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)
    r = r.astype(int)

    tbin = np.bincount(r.ravel(), data.ravel())
    nr = np.bincount(r.ravel())
    radialprofile = tbin / nr
    return radialprofile


@export
def gradient_map(image):
    """Computes gradients of images with the sobel transform.

    :param image: 2d numpy array
    :return: array of same size as input, with gradients between neighboring pixels
    """
    from skimage import filters

    return filters.sobel(image)



from lenstronomy.Data.psf import PSF
import lenstronomy.Util.util as util
import lenstronomy.Util.image_util as image_util
import lenstronomy.Util.kernel_util as kernel_util
import lenstronomy.Util.mask_util as mask_util

import numpy as np
import copy
from scipy import ndimage

__all__ = ["PsfFitting"]


class PsfFitting(object):
    """Class to find subsequently a better psf The method make use of a model and
    subtracts all the non-point source components of the model from the data. If the
    model is sufficient, then the data will be a (better) representation of the actual
    PSF. The method cuts out those point sources and combines them to update the
    estimate of the PSF. This is done in an iterative procedure as the model components
    of the extended features is PSF-dependent (hopefully not too much).

    Various options can be chosen. There is no guarantee that the method works for specific data and models.

    'stacking_method': 'median', 'mean'; the different estimates of the PSF are stacked and combined together.
    The choices are:

    - 'mean': mean of pixel values as the estimator (not robust to outliers)
    - 'median': median of pixel values as the estimator (outlier rejection robust but needs >2 point sources in the data

    'block_center_neighbour': angle, radius of neighbouring point sources around their centers the estimates is ignored.
        Default is zero, meaning a not optimal subtraction of the neighbouring point sources might contaminate the
        estimate.

    'keep_error_map': bool, if True, does not replace the error term associated with the PSF estimate.
        If false, re-estimates the variance between the PSF estimates.

    'psf_symmetry': number of rotational invariant symmetries in the estimated PSF.
        =1 mean no additional symmetries. =4 means 90 deg symmetry. This is enforced by a rotatioanl stack according to
        the symmetry specified. These additional imposed symmetries can help stabelize the PSF estimate when there are
        limited constraints/number of point sources in the image.

    The procedure only requires and changes the 'point_source_kernel' in the PSF() class and the 'psf_error_map'.
    Any previously set subgrid kernels or pixel_kernels are removed and constructed from the 'point_source_kernel'.
    """

    def __init__(self, image_model_class):
        """

        :param image_model_class: ImageModel class instance
        """
        self._image_model_class = image_model_class

    @staticmethod
    def calc_cornermask(kernelsize, psf_symmetry):
        """Calculate the completeness numerically when rotational symmetry is imposed.
        This is the simplest 'mask' which throws away anywhere the rotations are not
        fully complete ->e.g. in the corners. This ONLY accounts for information loss in
        corners, not due e.g. to losses at the edges of the images.

        :param kernelsize: int, size of kernel array
        :param psf_symmetry: int, the symmetry being imposed on the data
        :return: mask showing where the psf with symmetry n is incomplete due to
            rotation.
        """
        angle = 360.0 / psf_symmetry

        ones_im = np.ones((kernelsize, kernelsize))
        corner_norm_array = np.zeros((psf_symmetry, kernelsize, kernelsize))
        for k in range(psf_symmetry):
            ones_im_rotated = image_util.rotateImage(ones_im, angle * k)
            ones_im_rotated[ones_im_rotated > 1] = 1
            corner_norm_array[k, :, :] = ones_im_rotated
        total_corner_norm = np.sum(corner_norm_array, axis=0)
        mask = total_corner_norm < total_corner_norm.max()
        return mask

    def update_iterative(
        self,
        kwargs_psf,
        kwargs_params,
        num_iter=10,
        keep_psf_error_map=True,
        no_break=True,
        verbose=True,
        **kwargs_psf_update
    ):
        """

        :param kwargs_psf: keyword arguments to construct the PSF() class
        :param kwargs_params: keyword arguments of the parameters of the model components (e.g. 'kwargs_lens' etc)
        :param num_iter: number of iterations in the PSF fitting and image fitting process
        :param keep_psf_error_map: boolean, if True keeps previous psf_error_map
        :param no_break: boolean, if True, runs until the end regardless of the next step getting worse, and then
         reads out the overall best fit
        :param verbose: print statements informing about progress of iterative procedure
        :param kwargs_psf_update: keyword arguments providing the settings for a single iteration of the PSF, as being
         passed to update_psf() method
        :return: keyword argument of PSF constructor for PSF() class with updated PSF
        """
        self._image_model_class.PointSource.set_save_cache(True)
        if "kernel_point_source_init" not in kwargs_psf:
            kernel_point_source_init = copy.deepcopy(kwargs_psf["kernel_point_source"])
        else:
            kernel_point_source_init = kwargs_psf["kernel_point_source_init"]
        kwargs_psf_new = copy.deepcopy(kwargs_psf)
        kwargs_psf_final = copy.deepcopy(kwargs_psf)
        if "psf_error_map" in kwargs_psf:
            error_map_final = kwargs_psf["psf_error_map"]
        else:
            error_map_final = np.zeros_like(kernel_point_source_init)
        error_map_init = copy.deepcopy(error_map_final)
        psf_class = PSF(**kwargs_psf)
        self._image_model_class.update_psf(psf_class)
        logL_before = self._image_model_class.likelihood_data_given_model(
            **kwargs_params
        )
        logL_best = copy.deepcopy(logL_before)
        i_best = 0

        corner_mask = None
        if "corner_symmetry" in kwargs_psf_update.keys():
            if type(kwargs_psf_update["corner_symmetry"]) == int:
                psf_symmetry = kwargs_psf_update["psf_symmetry"]
                kernel_size = len(kwargs_psf["kernel_point_source"])
                corner_mask = self.calc_cornermask(kernel_size, psf_symmetry)

        for i in range(num_iter):
            kwargs_psf_new, logL_after, error_map = self.update_psf(
                kwargs_psf_new, kwargs_params, corner_mask, **kwargs_psf_update
            )

            if logL_after > logL_best:
                kwargs_psf_final = copy.deepcopy(kwargs_psf_new)
                error_map_final = copy.deepcopy(error_map)
                logL_best = logL_after
                i_best = i + 1
            else:
                if not no_break:
                    if verbose:
                        print(
                            "iterative PSF reconstruction makes reconstruction worse in step %s - aborted"
                            % i
                        )
                    break
        if verbose is True:
            print("iteration of step %s gave best reconstruction." % i_best)
            print(
                "log likelihood before: %s and log likelihood after: %s"
                % (logL_before, logL_best)
            )
        if keep_psf_error_map is True:
            kwargs_psf_final["psf_error_map"] = error_map_init
        else:
            kwargs_psf_final["psf_error_map"] = error_map_final
        kwargs_psf_final["kernel_point_source_init"] = kernel_point_source_init
        return kwargs_psf_final

    def update_psf(
        self,
        kwargs_psf,
        kwargs_params,
        corner_mask=None,
        stacking_method="median",
        psf_symmetry=1,
        psf_iter_factor=0.2,
        block_center_neighbour=0,
        error_map_radius=None,
        block_center_neighbour_error_map=None,
        new_procedure=True,
        corner_symmetry=None,
    ):
        """

        :param kwargs_psf: keyword arguments to construct the PSF() class
        :param kwargs_params: keyword arguments of the parameters of the model components (e.g. 'kwargs_lens' etc)
        :param stacking_method: 'median', 'mean'; the different estimates of the PSF are stacked and combined together.
         The choices are:
         'mean': mean of pixel values as the estimator (not robust to outliers)
         'median': median of pixel values as the estimator (outlier rejection robust but needs >2 point sources in the
         data
        :param psf_symmetry: number of rotational invariant symmetries in the estimated PSF.
         =1 mean no additional symmetries. =4 means 90 deg symmetry. This is enforced by a rotatioanl stack according to
         the symmetry specified. These additional imposed symmetries can help stabelize the PSF estimate when there are
         limited constraints/number of point sources in the image.
        :param psf_iter_factor: factor in (0, 1] of ratio of old vs new PSF in the update in the iteration.
        :param block_center_neighbour: angle, radius of neighbouring point sources around their centers the estimates
         is ignored. Default is zero, meaning a not optimal subtraction of the neighbouring point sources might
         contaminate the estimate.
        :param block_center_neighbour_error_map: angle, radius of neighbouring point sources around their centers the
         estimates of the ERROR MAP is ignored. If None, then the value of block_center_neighbour is used (recommended)
        :param error_map_radius: float, radius (in arc seconds) of the outermost error in the PSF estimate
         (e.g. to avoid double counting of overlapping PSF errors), if None, all of the pixels are considered
         (unless blocked through other means)
        :param new_procedure: boolean, uses post lenstronomy 1.9.2 procedure which is more optimal for super-sampled
         PSF's
        :param corner_mask: a mask which tracks completeness due to non 90 degree rotation for PSF symmetry.
         computed before this function to save time.
        :param corner_symmetry: int, if the imposed symmetry is an odd number, the edges of the reconstructed PSF in its default form will be
         clipped at the corners. corner_symmetry
         1) tracks where the residuals are being clipped by the imposed symmetry and then
         2) creates a psf with no symmetry
         3) adds the corner_symmetry psf (which has information at the corners) to the odd symmetry PSF, in the regions
         where the odd-symmetry PSF does not have information
        :return: kwargs_psf_new, logL_after, error_map
        """
        if block_center_neighbour_error_map is None:
            block_center_neighbour_error_map = block_center_neighbour
        psf_class = PSF(**kwargs_psf)
        kwargs_psf_copy = copy.deepcopy(kwargs_psf)

        point_source_supersampling_factor = kwargs_psf_copy.get(
            "point_source_supersampling_factor", 1
        )

        kwargs_psf_new = {
            "psf_type": "PIXEL",
            "kernel_point_source": kwargs_psf_copy["kernel_point_source"],
            "point_source_supersampling_factor": point_source_supersampling_factor,
            "psf_error_map": kwargs_psf_copy.get("psf_error_map", None),
        }
        # if 'psf_error_map' in kwargs_psf_copy:
        #    kwargs_psf_new['psf_error_map'] = kwargs_psf_copy['psf_error_map'] / 10
        self._image_model_class.update_psf(PSF(**kwargs_psf_new))

        (
            model,
            error_map_image,
            cov_param,
            param,
        ) = self._image_model_class.image_linear_solve(**kwargs_params)
        kwargs_ps = kwargs_params.get("kwargs_ps", None)
        kwargs_lens = kwargs_params.get("kwargs_lens", None)
        (
            ra_image,
            dec_image,
            point_amp,
        ) = self._image_model_class.PointSource.point_source_list(
            kwargs_ps, kwargs_lens
        )
        x_, y_ = self._image_model_class.Data.map_coord2pix(ra_image, dec_image)
        kernel_old = psf_class.kernel_point_source
        kernel_size = len(kernel_old)

        if not new_procedure:
            image_single_point_source_list = self.image_single_point_source(
                self._image_model_class, kwargs_params
            )
            star_cutout_list = self.point_like_source_cutouts(
                x_pos=x_,
                y_pos=y_,
                image_list=image_single_point_source_list,
                cutout_size=kernel_size,
            )
            psf_kernel_list = self.cutout_psf(
                ra_image,
                dec_image,
                x_,
                y_,
                image_single_point_source_list,
                kernel_size,
                kernel_old,
                block_center_neighbour=block_center_neighbour,
            )

            kernel_new = self.combine_psf(
                psf_kernel_list,
                kernel_old,
                factor=psf_iter_factor,
                stacking_option=stacking_method,
                symmetry=psf_symmetry,
            )
            kernel_new = kernel_util.cut_psf(kernel_new, psf_size=kernel_size)
            error_map = self.error_map_estimate(
                kernel_new,
                star_cutout_list,
                point_amp,
                x_,
                y_,
                error_map_radius=error_map_radius,
                block_center_neighbour=block_center_neighbour_error_map,
            )

            if point_source_supersampling_factor > 1:
                # The current version of using a super-sampled PSF in the iterative reconstruction is to first
                # constrain a down-sampled version and then in a second step perform a super-sampling of it. This is not
                # optimal and should be changed in the future that the corrections of the super-sampled version is done
                # rather than constraining a totally new PSF first
                kernel_new = kernel_util.subgrid_kernel(
                    kernel_new,
                    subgrid_res=point_source_supersampling_factor,
                    odd=True,
                    num_iter=10,
                )
                # chop edges
                n_kernel = len(kwargs_psf["kernel_point_source"])
                kernel_new = kernel_util.cut_psf(kernel_new, psf_size=n_kernel)

        else:
            kernel_old_high_res = psf_class.kernel_point_source_supersampled(
                supersampling_factor=point_source_supersampling_factor
            )
            kernel_size_high = len(kernel_old_high_res)
            data = self._image_model_class.Data.data
            residuals = data - model

            psf_kernel_list = self.psf_estimate_individual(
                ra_image,
                dec_image,
                point_amp,
                residuals,
                cutout_size=kernel_size,
                kernel_guess=kernel_old_high_res,
                supersampling_factor=point_source_supersampling_factor,
                block_center_neighbour=block_center_neighbour,
            )

            kernel_new = self.combine_psf(
                psf_kernel_list,
                kernel_old_high_res,
                factor=psf_iter_factor,
                stacking_option=stacking_method,
                symmetry=psf_symmetry,
                corner_symmetry=corner_symmetry,
                corner_mask=corner_mask,
            )
            kernel_new = kernel_util.cut_psf(kernel_new, psf_size=kernel_size_high)

            # resize kernel for error_map estimate
            # kernel_new_low = kernel_util.degrade_kernel(kernel_new, point_source_supersampling_factor)
            # compute error map on pixel level
            error_map = self.error_map_estimate_new(
                kernel_new,
                psf_kernel_list,
                ra_image,
                dec_image,
                point_amp,
                point_source_supersampling_factor,
                error_map_radius=error_map_radius,
            )

        kwargs_psf_new["kernel_point_source"] = kernel_new
        # if 'psf_error_map' in kwargs_psf_new:
        #    kwargs_psf_new['psf_error_map'] *= 10
        self._image_model_class.update_psf(PSF(**kwargs_psf_new))
        logL_after = self._image_model_class.likelihood_data_given_model(
            **kwargs_params
        )
        return kwargs_psf_new, logL_after, error_map

    def image_single_point_source(self, image_model_class, kwargs_params):
        """Return model without including the point source contributions as a list (for
        each point source individually)

        :param image_model_class: ImageModel class instance
        :param kwargs_params: keyword arguments of model component keyword argument
            lists
        :return: list of images with point source isolated
        """
        # reconstructed model with given psf
        model, error_map, cov_param, param = image_model_class.image_linear_solve(
            **kwargs_params
        )
        data = image_model_class.Data.data
        mask = image_model_class.likelihood_mask
        kwargs_ps = kwargs_params.get("kwargs_ps", None)
        kwargs_lens = kwargs_params.get("kwargs_lens", None)
        point_source_list = self._point_sources_list(
            image_model_class, kwargs_ps, kwargs_lens
        )
        n = len(point_source_list)
        model_single_source_list = []
        for i in range(n):
            model_single_source = (data - model + point_source_list[i]) * mask
            model_single_source_list.append(model_single_source)
        return model_single_source_list

    @staticmethod
    def _point_sources_list(image_model_class, kwargs_ps, kwargs_lens, k=None):
        """

        :param kwargs_ps:
        :return: list of images containing only single point sources
        """
        point_list = []
        (
            ra_array,
            dec_array,
            amp_array,
        ) = image_model_class.PointSource.point_source_list(kwargs_ps, kwargs_lens, k=k)
        for i in range(len(ra_array)):
            point_source = image_model_class.ImageNumerics.point_source_rendering(
                [ra_array[i]], [dec_array[i]], [amp_array[i]]
            )
            point_list.append(point_source)
        return point_list

    def cutout_psf(
        self,
        ra_image,
        dec_image,
        x,
        y,
        image_list,
        kernel_size,
        kernel_init,
        block_center_neighbour=0,
    ):
        """

        :param ra_image: coordinate array of images in angles
        :param dec_image: coordinate array of images in angles
        :param x: image position array in x-pixel
        :param y: image position array in y-pixel
        :param image_list: list of images (i.e. data - all models subtracted, except a single point source)
        :param kernel_size: width in pixel of the kernel
        :param kernel_init: initial guess of kernel (pixels that are masked are replaced by those values)
        :param block_center_neighbour: angle, radius of neighbouring point sources around their centers the estimates
         is ignored. Default is zero, meaning a not optimal subtraction of the neighbouring point sources might
         contaminate the estimate.
        :return: list of de-shifted kernel estimates
        """
        mask = self._image_model_class.likelihood_mask
        ra_grid, dec_grid = self._image_model_class.Data.pixel_coordinates
        ra_grid = util.image2array(ra_grid)
        dec_grid = util.image2array(dec_grid)
        radius = block_center_neighbour

        kernel_list = []
        for l in range(len(x)):
            mask_point_source = self.mask_point_source(
                ra_image, dec_image, ra_grid, dec_grid, radius, i=l
            )
            mask_i = mask * mask_point_source
            kernel_deshifted = self.cutout_psf_single(
                x[l], y[l], image_list[l], mask_i, kernel_size, kernel_init
            )
            kernel_list.append(kernel_deshifted)
        return kernel_list

    def psf_estimate_individual(
        self,
        ra_image,
        dec_image,
        point_amp,
        residuals,
        cutout_size,
        kernel_guess,
        supersampling_factor,
        block_center_neighbour,
    ):
        """

        :param ra_image: list; position in angular units of the image
        :param dec_image: list; position in angular units of the image
        :param point_amp: list of model amplitudes of point sources
        :param residuals: data - model
        :param cutout_size: pixel size of cutout around single star/quasar to be considered for the psf reconstruction
        :param kernel_guess: initial guess of super-sampled PSF
        :param supersampling_factor: int, super-sampling factor
        :param block_center_neighbour:
        :return: list of best-guess PSF's for each star based on the residual patterns
        """
        mask = self._image_model_class.likelihood_mask
        ra_grid, dec_grid = self._image_model_class.Data.pixel_coordinates
        ra_grid = util.image2array(ra_grid)
        dec_grid = util.image2array(dec_grid)
        radius = block_center_neighbour
        x_, y_ = self._image_model_class.Data.map_coord2pix(ra_image, dec_image)

        kernel_list = []
        for l in range(len(ra_image)):
            mask_point_source = self.mask_point_source(
                ra_image, dec_image, ra_grid, dec_grid, radius, i=l
            )
            mask_i = mask * mask_point_source

            # cutout residuals
            x_int = int(round(x_[l]))
            y_int = int(round(y_[l]))
            residual_cutout = kernel_util.cutout_source(
                x_int, y_int, residuals, cutout_size + 2, shift=False
            )
            # cutout the mask
            mask_cutout = kernel_util.cutout_source(
                x_int, y_int, mask_i, cutout_size + 2, shift=False
            )
            # apply mask
            residual_cutout_mask = residual_cutout * mask_cutout
            # re-scale residuals with point source brightness
            residual_cutout_mask /= point_amp[l]
            # enlarge residuals by super-sampling factor
            residual_cutout_mask = residual_cutout_mask.repeat(
                supersampling_factor, axis=0
            ).repeat(supersampling_factor, axis=1)

            # inverse shift residuals
            shift_x = (x_int - x_[l]) * supersampling_factor
            shift_y = (y_int - y_[l]) * supersampling_factor
            # for odd number super-sampling
            if supersampling_factor % 2 == 1:
                residuals_shifted = ndimage.shift(
                    residual_cutout_mask, shift=[shift_y, shift_x], order=1
                )

            else:
                # for even number super-sampling half a super-sampled pixel offset needs to be performed
                residuals_shifted = ndimage.shift(
                    residual_cutout_mask, shift=[shift_y - 0.5, shift_x - 0.5], order=1
                )
                # and the last column and row need to be removed
                residuals_shifted = residuals_shifted[:-1, :-1]

            # re-size shift residuals
            psf_size = len(kernel_guess)
            residuals_shifted = image_util.cut_edges(residuals_shifted, psf_size)

            # normalize residuals
            correction = residuals_shifted - np.mean(residuals_shifted)
            # correct old PSF with inverse shifted residuals
            kernel_new = kernel_guess + correction
            kernel_list.append(kernel_new)
        return kernel_list

    @staticmethod
    def point_like_source_cutouts(x_pos, y_pos, image_list, cutout_size):
        """Cutouts of point-like objects.

        :param x_pos: list of image positions in pixel units
        :param y_pos: list of image position in pixel units
        :param image_list: list of 2d numpy arrays with cleaned images, with all
            contaminating sources removed except the point-like object to be cut out.
        :param cutout_size: odd integer, size of cutout.
        :return: list of cutouts
        """

        star_cutout_list = []
        for l in range(len(x_pos)):
            x_int = int(round(x_pos[l]))
            y_int = int(round(y_pos[l]))
            star_cutout = kernel_util.cutout_source(
                x_int, y_int, image_list[l], cutout_size, shift=False
            )
            star_cutout_list.append(star_cutout)
        return star_cutout_list

    @staticmethod
    def cutout_psf_single(x, y, image, mask, kernel_size, kernel_init):
        """

        :param x: x-coordinate of point source
        :param y: y-coordinate of point source
        :param image: image (i.e. data - all models subtracted, except a single point source)
        :param mask: mask of pixels in the image not to be considered in the PSF estimate (being replaced by kernel_init)
        :param kernel_size: width in pixel of the kernel
        :param kernel_init: initial guess of kernel (pixels that are masked are replaced by those values)
        :return: estimate of the PSF based on the image and position of the point source
        """
        # cutout the star
        x_int = int(round(x))
        y_int = int(round(y))
        star_cutout = kernel_util.cutout_source(
            x_int, y_int, image, kernel_size + 2, shift=False
        )
        # cutout the mask
        mask_cutout = kernel_util.cutout_source(
            x_int, y_int, mask, kernel_size + 2, shift=False
        )
        # enlarge the initial PSF kernel to the new cutout size
        kernel_enlarged = np.zeros((kernel_size + 2, kernel_size + 2))
        kernel_enlarged[1:-1, 1:-1] = kernel_init
        # shift the initial kernel to the shift of the star
        shift_x = x_int - x
        shift_y = y_int - y
        kernel_shifted = ndimage.shift(
            kernel_enlarged, shift=[-shift_y, -shift_x], order=1
        )
        # compute normalization of masked and unmasked region of the shifted kernel
        # norm_masked = np.sum(kernel_shifted[mask_i == 0])
        norm_unmasked = np.sum(kernel_shifted[mask_cutout == 1])
        # normalize star within the unmasked region to the norm of the initial kernel of the same region
        star_cutout /= np.sum(star_cutout[mask_cutout == 1]) * norm_unmasked
        # replace mask with shifted initial kernel (+2 size)
        star_cutout[mask_cutout == 0] = kernel_shifted[mask_cutout == 0]
        star_cutout[star_cutout < 0] = 0
        # de-shift kernel
        kernel_deshifted = kernel_util.de_shift_kernel(
            star_cutout, shift_x, shift_y, iterations=20, fractional_step_size=0.1
        )
        # re-size kernel
        kernel_deshifted = image_util.cut_edges(kernel_deshifted, kernel_size)
        # re-normalize kernel again
        kernel_deshifted = kernel_util.kernel_norm(kernel_deshifted)
        return kernel_deshifted

    @staticmethod
    ####Correct this based on Maverick Oh's note about lack of flux conservation.
    def combine_psf(
        kernel_list_new,
        kernel_old,
        factor=1.0,
        stacking_option="median",
        symmetry=1,
        corner_symmetry=None,
        corner_mask=None,
    ):
        ## TODO: Account for image edges/masked pixels.
        """Updates psf estimate based on old kernel and several new estimates.

        :param kernel_list_new: list of new PSF kernels estimated from the point sources
            in the image (un-normalized)
        :param kernel_old: old PSF kernel
        :param factor: weight of updated estimate based on new and old estimate,
            factor=1 means new estimate, factor=0 means old estimate
        :param stacking_option: option of stacking, mean or median
        :param symmetry: imposed symmetry of PSF estimate
        :param corner_symmetry: int, if the imposed symmetry is an odd number, the edges
            of the reconstructed PSF in its default form will be clipped at the corners.
            corner_symmetry 1) tracks where the residuals are being clipped by the
            imposed symmetry and then 2) creates a psf with symmetry=corner symmetry
            which is either 1 or 360/symm = n*90. (e.g for a symmetry 6 psf you could
            use symmetry 2 in the corners). 3) adds the corner_symmetry psf (which has
            information at the corners) to the odd symmetry PSF, in the regions where
            the odd-symmetry PSF does not have complete information.
        :return: updated PSF estimate
        """
        ## keep_corners is a boolean which tracks whether to calc PSF separately in the corners for odd symmetry rotations.
        keep_corners = type(corner_symmetry) == int
        n = int(len(kernel_list_new) * symmetry)
        angle = 360.0 / symmetry
        kernelsize = len(kernel_old)

        kernel_list = np.zeros((n, kernelsize, kernelsize))

        if keep_corners:
            n_corners = int(len(kernel_list_new * corner_symmetry))
            angle_corner = 360.0 / corner_symmetry
            corner_kernel_array = np.zeros((n_corners, kernelsize, kernelsize))

        i = 0
        m = 0
        for kernel_new in kernel_list_new:
            ##normalize each residual kernel one time at the start, before rotations clip them.
            kernel_new = kernel_util.kernel_norm(kernel_new)
            for k in range(symmetry):
                kernel_rotated = image_util.rotateImage(kernel_new, angle * k)
                kernel_list[i, :, :] = kernel_rotated
                i += 1

            ###do a rotation for the corner part of the data (i.e. if symmetry is 2 or 4).
            if keep_corners:
                for j in range(corner_symmetry):
                    corner_kernel_rotated = image_util.rotateImage(
                        kernel_new, angle_corner * j
                    )
                    corner_kernel_array[m, :, :] = corner_kernel_rotated
                    m += 1

        if stacking_option == "median":
            ##previous version took the median including the old kernel (extended kernel list with rotated old kernel)
            # Now remove that and use the weighting later for stabilization
            kernel_new = np.median(kernel_list, axis=0)
            if keep_corners:
                kernel_new_corners = np.median(corner_kernel_array, axis=0)

        elif stacking_option == "mean":
            ##previous version took the mean including the old kernel. Now remove that and use the weighting later for stabilization
            kernel_new = np.mean(kernel_list, axis=0)
            if keep_corners:
                kernel_new_corners = np.mean(corner_kernel_array, axis=0)

        else:
            raise ValueError(
                " stack_option must be 'median' or 'mean', %s is not supported."
                % stacking_option
            )

        ###calculate the completeness for the main rotational symmetry--> anywhere this is not 1, only use the 'corners'
        # kernel future improvement: do a weighted median/mean based on this normalization.
        kernel_new = np.nan_to_num(kernel_new)
        kernel_new[kernel_new < 0] = 0

        if keep_corners:
            kernel_new_corners = np.nan_to_num(kernel_new_corners)
            kernel_new_corners[kernel_new_corners < 0] = 0
            ##anywhere you didn't have complete data for n_symmetry exposures, then substitute the corners kernel.
            kernel_new[corner_mask] = kernel_new_corners[corner_mask]

        ###just in case the old kernel is not normalized. Probably want to do this earlier elsewhere, but @simon can let me know if this is necessary here.
        kernel_old = kernel_util.kernel_norm(kernel_old)
        kernel_new = kernel_util.kernel_norm(kernel_new)
        kernel_return = factor * kernel_new + (1.0 - factor) * kernel_old
        return kernel_return

    def error_map_estimate_new(
        self,
        psf_kernel,
        psf_kernel_list,
        ra_image,
        dec_image,
        point_amp,
        supersampling_factor,
        error_map_radius=None,
    ):
        """Relative uncertainty in the psf model (in quadrature) per pixel based on
        residuals achieved in the image.

        :param psf_kernel: PSF kernel (super-sampled)
        :param psf_kernel_list: list of individual best PSF kernel estimates
        :param ra_image: image positions in angles
        :param dec_image: image positions in angles
        :param point_amp: image amplitude
        :param supersampling_factor: super-sampling factor
        :param error_map_radius: radius (in angle) to cut the error map
        :return: psf error map such that square of the uncertainty gets boosted by
            error_map * (psf * amp)**2
        """
        kernel_low = kernel_util.degrade_kernel(psf_kernel, supersampling_factor)
        error_map_list = np.zeros(
            (len(psf_kernel_list), len(kernel_low), len(kernel_low))
        )
        x_pos, y_pos = self._image_model_class.Data.map_coord2pix(ra_image, dec_image)

        for i, psf_kernel_i in enumerate(psf_kernel_list):
            kernel_low_i = kernel_util.degrade_kernel(
                psf_kernel_i, supersampling_factor
            )

            x, y, amp_i = x_pos[i], y_pos[i], point_amp[i]
            x_int = int(round(x))
            y_int = int(round(y))

            C_D_cutout = kernel_util.cutout_source(
                x_int,
                y_int,
                self._image_model_class.Data.C_D,
                len(kernel_low_i),
                shift=False,
            )
            residuals_i = np.abs(kernel_low - kernel_low_i)
            residuals_i -= np.sqrt(C_D_cutout) / amp_i
            residuals_i[residuals_i < 0] = 0
            error_map_list[i, :, :] = residuals_i**2

        error_map = np.median(error_map_list, axis=0)
        error_map[kernel_low > 0] /= kernel_low[kernel_low > 0] ** 2
        error_map = np.nan_to_num(error_map)
        error_map[error_map > 1] = 1  # cap on error to be the same

        # mask the error map outside a certain radius (can avoid double counting of errors when map is overlapping
        if error_map_radius is not None:
            pixel_scale = self._image_model_class.Data.pixel_width
            x_grid, y_grid = util.make_grid(numPix=len(error_map), deltapix=pixel_scale)
            mask = mask_util.mask_azimuthal(
                x_grid, y_grid, center_x=0, center_y=0, r=error_map_radius
            )
            error_map *= util.array2image(mask)
        return error_map

    def error_map_estimate(
        self,
        kernel,
        star_cutout_list,
        amp,
        x_pos,
        y_pos,
        error_map_radius=None,
        block_center_neighbour=0,
    ):
        """Provides a psf_error_map based on the goodness of fit of the given PSF kernel
        on the point source cutouts, their estimated amplitudes and positions.

        :param kernel: PSF kernel
        :param star_cutout_list: list of 2d arrays of cutouts of the point sources with
            all other model components subtracted
        :param amp: list of amplitudes of the estimated PSF kernel
        :param x_pos: pixel position (in original data unit, not in cutout) of the point
            sources (same order as amp and star cutouts)
        :param y_pos: pixel position (in original data unit, not in cutout) of the point
            sources (same order as amp and star cutouts)
        :param error_map_radius: float, radius (in arc seconds) of the outermost error
            in the PSF estimate (e.g. to avoid double counting of overlapping PSF erros)
        :param block_center_neighbour: angle, radius of neighbouring point sources
            around their centers the estimates is ignored. Default is zero, meaning a
            not optimal subtraction of the neighbouring point sources might contaminate
            the estimate.
        :return: relative uncertainty in the psf model (in quadrature) per pixel based
            on residuals achieved in the image
        """
        error_map_list = np.zeros((len(star_cutout_list), len(kernel), len(kernel)))
        mask_list = np.zeros((len(star_cutout_list), len(kernel), len(kernel)))
        ra_grid, dec_grid = self._image_model_class.Data.pixel_coordinates
        ra_grid = util.image2array(ra_grid)
        dec_grid = util.image2array(dec_grid)
        mask = self._image_model_class.likelihood_mask
        for i, star in enumerate(star_cutout_list):
            x, y, amp_i = x_pos[i], y_pos[i], amp[i]
            # shift kernel
            x_int = int(round(x))
            y_int = int(round(y))
            shift_x = x_int - x
            shift_y = y_int - y
            kernel_shifted = ndimage.shift(kernel, shift=[-shift_y, -shift_x], order=1)
            # multiply kernel with amplitude
            model = kernel_shifted * amp_i
            # compute residuals
            residual = np.abs(star - model)
            # subtract background and Poisson noise residuals
            C_D_cutout = kernel_util.cutout_source(
                x_int, y_int, self._image_model_class.Data.C_D, len(star), shift=False
            )
            # block neighbor points in error estimate
            mask_point_source = self.mask_point_source(
                x_pos, y_pos, ra_grid, dec_grid, radius=block_center_neighbour, i=i
            )
            mask_i = mask * mask_point_source
            mask_i = kernel_util.cutout_source(
                x_int, y_int, mask_i, len(star), shift=False
            )
            residual -= np.sqrt(C_D_cutout)
            residual[residual < 0] = 0
            # estimate relative error per star
            residual /= amp_i
            error_map_list[i, :, :] = residual**2 * mask_i
            mask_list[i, :, :] = mask_i
        # take median absolute error for each pixel
        # TODO: only for pixels that are not masked
        error_map = np.median(error_map_list, axis=0)
        error_map[kernel > 0] /= kernel[kernel > 0] ** 2
        error_map = np.nan_to_num(error_map)
        error_map[error_map > 1] = 1  # cap on error to be the same

        # mask the error map outside a certain radius (can avoid double counting of errors when map is overlapping
        if error_map_radius is not None:
            pixel_scale = self._image_model_class.Data.pixel_width
            x_grid, y_grid = util.make_grid(numPix=len(error_map), deltapix=pixel_scale)
            mask = mask_util.mask_azimuthal(
                x_grid, y_grid, center_x=0, center_y=0, r=error_map_radius
            )
            error_map *= util.array2image(mask)
        return error_map

    @staticmethod
    def mask_point_source(x_pos, y_pos, x_grid, y_grid, radius, i=0):
        """

        :param x_pos: x-position of list of point sources
        :param y_pos: y-position of list of point sources
        :param x_grid: x-coordinates of grid
        :param y_grid: y-coordinates of grid
        :param i: index of point source not to mask out
        :param radius: radius to mask out other point sources
        :return: a mask of the size of the image with cutouts around the position
        """
        mask = np.ones_like(x_grid)
        for k in range(len(x_pos)):
            if k != i:
                mask_point = 1 - mask_util.mask_azimuthal(
                    x_grid, y_grid, x_pos[k], y_pos[k], radius
                )
                mask *= mask_point
        return util.array2image(mask)



__author__ = "sibirrer"


import numpy as np
import pytest
import numpy.testing as npt
import lenstronomy.Util.param_util as param_util

try:
    import fastell4py

    fastell4py_bool = True
except:
    print(
        "Warning: fastell4py not available, tests will be trivially fulfilled without giving the right answer!"
    )
    fastell4py_bool = False


class TestSPEMD(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.pemd import PEMD
        from lenstronomy.LensModel.Profiles.spep import SPEP

        self.PEMD = PEMD(suppress_fastell=True)
        self.SPEP = SPEP()

    def test_function(self):
        phi_E = 1.0
        gamma = 1.9
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0])
        y = np.array([2])
        a = np.zeros_like(x)
        values = self.PEMD.function(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(values[0], 2.1571106351401803, decimal=5)
        else:
            assert values == 0
        a += values
        x = np.array(1.0)
        y = np.array(2.0)
        a = np.zeros_like(x)
        values = self.PEMD.function(x, y, phi_E, gamma, e1, e2)
        print(x, values)
        a += values
        if fastell4py_bool:
            npt.assert_almost_equal(values, 2.1571106351401803, decimal=5)
        else:
            assert values == 0
        assert type(x) == type(values)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.PEMD.function(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(values[0], 2.180188584782964, decimal=7)
            npt.assert_almost_equal(values[1], 3.2097137160951874, decimal=7)
            npt.assert_almost_equal(values[2], 4.3109976673748, decimal=7)
        else:
            npt.assert_almost_equal(values[0], 0, decimal=7)
            npt.assert_almost_equal(values[1], 0, decimal=7)
            npt.assert_almost_equal(values[2], 0, decimal=7)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        gamma = 1.9
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.PEMD.derivatives(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_x[0], 0.46664118422711387, decimal=7)
            npt.assert_almost_equal(f_y[0], 0.9530892465981603, decimal=7)
        else:
            npt.assert_almost_equal(f_x[0], 0, decimal=7)
            npt.assert_almost_equal(f_y[0], 0, decimal=7)

        x = np.array([1.0, 3, 4])
        y = np.array([2.0, 1, 1])
        a = np.zeros_like(x)
        values = self.PEMD.derivatives(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(values[0][0], 0.46664118422711387, decimal=7)
            npt.assert_almost_equal(values[1][0], 0.9530892465981603, decimal=7)
            npt.assert_almost_equal(values[0][1], 1.0722265330847958, decimal=7)
            npt.assert_almost_equal(values[1][1], 0.3140067377020791, decimal=7)
        else:
            npt.assert_almost_equal(values[0][0], 0, decimal=7)
            npt.assert_almost_equal(values[1][0], 0, decimal=7)
            npt.assert_almost_equal(values[0][1], 0, decimal=7)
            npt.assert_almost_equal(values[1][1], 0, decimal=7)
        a += values[0]
        x = 1.0
        y = 2.0
        phi_E = 1.0
        gamma = 1.9
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.PEMD.derivatives(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_x, 0.46664118422711387, decimal=7)
            npt.assert_almost_equal(f_y, 0.9530892465981603, decimal=7)
        else:
            npt.assert_almost_equal(f_x, 0, decimal=7)
            npt.assert_almost_equal(f_y, 0, decimal=7)
        x = 0.0
        y = 0.0
        f_x, f_y = self.PEMD.derivatives(x, y, phi_E, gamma, e1, e2)
        assert f_x == 0.0
        assert f_y == 0.0

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        gamma = 1.9
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.PEMD.hessian(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_xx, 0.4179041, decimal=7)
            npt.assert_almost_equal(f_yy, 0.1404714, decimal=7)
            npt.assert_almost_equal(f_xy, -0.1856134, decimal=7)
        else:
            npt.assert_almost_equal(f_xx, 0, decimal=7)
            npt.assert_almost_equal(f_yy, 0, decimal=7)
            npt.assert_almost_equal(f_xy, 0, decimal=7)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

        x = 1.0
        y = 2.0
        phi_E = 1.0
        gamma = 1.9
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        a = np.zeros_like(x)
        f_xx, f_xy, f_yx, f_yy = self.PEMD.hessian(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_xx, 0.41790408341142493, decimal=7)
            npt.assert_almost_equal(f_yy, 0.14047143086334482, decimal=7)
            npt.assert_almost_equal(f_xy, -0.1856133848300859, decimal=7)
        else:
            npt.assert_almost_equal(f_xx, 0, decimal=7)
            npt.assert_almost_equal(f_yy, 0, decimal=7)
            npt.assert_almost_equal(f_xy, 0, decimal=7)
        a += f_xx
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.PEMD.hessian(x, y, phi_E, gamma, e1, e2)
        print(values, "values")
        if fastell4py_bool:
            npt.assert_almost_equal(values[0][0], 0.41789957732890953, decimal=5)
            npt.assert_almost_equal(values[3][0], 0.14047593655054141, decimal=5)
            npt.assert_almost_equal(values[1][0], -0.18560737698052343, decimal=5)
            npt.assert_almost_equal(values[0][1], 0.068359818958208918, decimal=5)
            npt.assert_almost_equal(values[3][1], 0.32494089371516482, decimal=5)
            npt.assert_almost_equal(values[1][1], -0.097845438684594374, decimal=5)
        else:
            npt.assert_almost_equal(values[0][0], 0, decimal=7)

    def test_spep_spemd(self):
        x = np.array([1])
        y = np.array([0])
        phi_E = 1.0
        gamma = 2.0
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.PEMD.derivatives(x, y, phi_E, gamma, e1, e2)
        f_x_spep, f_y_spep = self.SPEP.derivatives(x, y, phi_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_x[0], f_x_spep[0], decimal=2)
        else:
            pass

        theta_E = 2.0
        gamma = 2.0
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.PEMD.derivatives(x, y, theta_E, gamma, e1, e2)
        f_x_spep, f_y_spep = self.SPEP.derivatives(x, y, theta_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_x[0], f_x_spep[0], decimal=2)
        else:
            pass

        theta_E = 2.0
        gamma = 1.7
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.PEMD.derivatives(x, y, theta_E, gamma, e1, e2)
        f_x_spep, f_y_spep = self.SPEP.derivatives(x, y, theta_E, gamma, e1, e2)
        if fastell4py_bool:
            npt.assert_almost_equal(f_x[0], f_x_spep[0], decimal=4)

    def test_bounds(self):
        from lenstronomy.LensModel.Profiles.spemd import SPEMD

        profile = SPEMD(suppress_fastell=True)
        compute_bool = profile._parameter_constraints(q_fastell=-1, gam=-1, s2=-1, q=-1)
        assert compute_bool is False

    def test_is_not_empty(self):
        func = self.PEMD.spemd_smooth.is_not_empty

        assert func(0.1, 0.2)
        assert func([0.1], [0.2])
        assert func((0.1, 0.3), (0.2, 0.4))
        assert func(np.array([0.1]), np.array([0.2]))
        assert not func([], [])
        assert not func(np.array([]), np.array([]))

    def test_density_lens(self):
        r = 1
        kwargs = {"theta_E": 1, "gamma": 2, "e1": 0, "e2": 0}
        rho = self.PEMD.density_lens(r, **kwargs)
        rho_spep = self.SPEP.density_lens(r, **kwargs)
        npt.assert_almost_equal(rho, rho_spep, decimal=7)


if __name__ == "__main__":
    pytest.main()
